import inspect
import json
import tempfile
from abc import abstractmethod
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import Literal, Optional, Type, Union

import numpy as np
import pytest
from hdmf_zarr import NWBZarrIO
from jsonschema.validators import Draft7Validator, validate
from numpy.testing import assert_array_equal
from pynwb import NWBHDF5IO
from pynwb.testing.mock.file import mock_NWBFile
from spikeinterface.core.testing import check_recordings_equal, check_sortings_equal

from neuroconv import BaseDataInterface, NWBConverter
from neuroconv.datainterfaces.ecephys.baserecordingextractorinterface import (
    BaseRecordingExtractorInterface,
)
from neuroconv.datainterfaces.ecephys.basesortingextractorinterface import (
    BaseSortingExtractorInterface,
)
from neuroconv.datainterfaces.ophys.baseimagingextractorinterface import (
    BaseImagingExtractorInterface,
)
from neuroconv.datainterfaces.ophys.basesegmentationextractorinterface import (
    BaseSegmentationExtractorInterface,
)
from neuroconv.tools.nwb_helpers import (
    configure_backend,
    get_default_backend_configuration,
)
from neuroconv.utils import _NWBMetaDataEncoder


class DataInterfaceTestMixin:
    """
    Generic class for testing DataInterfaces.


    Several of these tests are required to be run in a specific order. In this case,
    there is a `test_conversion_as_lone_interface` that calls the `check` functions in
    the appropriate order, after the `interface` has been created. Normally, you might
    expect the `interface` to be simply created in the `setUp` method, but this class
    allows you to specify multiple interface_kwargs.

    Class Attributes
    ----------------
    data_interface_cls : DataInterface
        class, not instance
    interface_kwargs : dict or list
        When it is a dictionary, take these as arguments to the constructor of the
        interface. When it is a list, each element of the list is a dictionary of
        arguments to the constructor. Each dictionary will be tested one at a time.
    save_directory : Path, optional
        Directory where test files should be saved.
    """

    data_interface_cls: Type[BaseDataInterface]
    interface_kwargs: dict
    save_directory: Path = Path(tempfile.mkdtemp())
    conversion_options: Optional[dict] = None
    maxDiff = None

    @pytest.fixture
    def setup_interface(self, request):
        """Add this as a fixture when you want freshly created interface in the test."""
        self.test_name: str = ""
        self.interface = self.data_interface_cls(**self.interface_kwargs)

        return self.interface, self.test_name

    @pytest.fixture(scope="class", autouse=True)
    def setup_default_conversion_options(self, request):
        cls = request.cls
        cls.conversion_options = cls.conversion_options or dict()
        return cls.conversion_options

    def test_source_schema_valid(self):
        schema = self.data_interface_cls.get_source_schema()
        Draft7Validator.check_schema(schema=schema)

    def check_conversion_options_schema_valid(self):
        schema = self.interface.get_conversion_options_schema()
        Draft7Validator.check_schema(schema=schema)

    def test_metadata_schema_valid(self, setup_interface):
        schema = self.interface.get_metadata_schema()
        Draft7Validator.check_schema(schema=schema)

    def check_metadata(self):
        # Validate metadata now happens on the class itself
        metadata = self.interface.get_metadata()
        self.check_extracted_metadata(metadata)

    def test_no_metadata_mutation(self, setup_interface):
        """Ensure the metadata object is not altered by `add_to_nwbfile` method."""

        nwbfile = mock_NWBFile()

        metadata = self.interface.get_metadata()
        metadata_before_add_method = deepcopy(metadata)

        self.interface.add_to_nwbfile(nwbfile=nwbfile, metadata=metadata, **self.conversion_options)
        assert metadata == metadata_before_add_method

    def check_run_conversion_with_backend_configuration(
        self, nwbfile_path: str, backend: Literal["hdf5", "zarr"] = "hdf5"
    ):
        metadata = self.interface.get_metadata()
        if "session_start_time" not in metadata["NWBFile"]:
            metadata["NWBFile"].update(session_start_time=datetime.now().astimezone())

        nwbfile = self.interface.create_nwbfile(metadata=metadata, **self.conversion_options)
        backend_configuration = self.interface.get_default_backend_configuration(nwbfile=nwbfile, backend=backend)
        self.interface.run_conversion(
            nwbfile_path=nwbfile_path,
            nwbfile=nwbfile,
            overwrite=True,
            metadata=metadata,
            backend_configuration=backend_configuration,
            **self.conversion_options,
        )

    def check_run_conversion_in_nwbconverter_with_backend(
        self, nwbfile_path: str, backend: Literal["hdf5", "zarr"] = "hdf5"
    ):
        class TestNWBConverter(NWBConverter):
            data_interface_classes = dict(Test=type(self.interface))

        source_data = dict(Test=self.interface_kwargs)
        converter = TestNWBConverter(source_data=source_data)

        metadata = converter.get_metadata()
        if "session_start_time" not in metadata["NWBFile"]:
            metadata["NWBFile"].update(session_start_time=datetime.now().astimezone())

        conversion_options = dict(Test=self.conversion_options)

        converter.run_conversion(
            nwbfile_path=nwbfile_path,
            overwrite=True,
            metadata=metadata,
            backend=backend,
            conversion_options=conversion_options,
        )

    def check_run_conversion_in_nwbconverter_with_backend_configuration(
        self, nwbfile_path: str, backend: Union["hdf5", "zarr"] = "hdf5"
    ):
        class TestNWBConverter(NWBConverter):
            data_interface_classes = dict(Test=type(self.interface))

        source_data = dict(Test=self.interface_kwargs)
        converter = TestNWBConverter(source_data=source_data)

        metadata = converter.get_metadata()
        if "session_start_time" not in metadata["NWBFile"]:
            metadata["NWBFile"].update(session_start_time=datetime.now().astimezone())

        conversion_options = dict(Test=self.conversion_options)

        nwbfile = converter.create_nwbfile(metadata=metadata, conversion_options=conversion_options)
        backend_configuration = converter.get_default_backend_configuration(nwbfile=nwbfile, backend=backend)
        converter.run_conversion(
            nwbfile_path=nwbfile_path,
            nwbfile=nwbfile,
            overwrite=True,
            metadata=metadata,
            backend_configuration=backend_configuration,
            conversion_options=conversion_options,
        )

    @abstractmethod
    def check_read_nwb(self, nwbfile_path: str):
        """Read the produced NWB file and compare it to the interface."""
        pass

    def check_extracted_metadata(self, metadata: dict):
        """Override this method to make assertions about specific extracted metadata values."""
        pass

    def run_custom_checks(self):
        """Override this in child classes to inject additional custom checks."""
        pass

    @pytest.mark.parametrize("backend", ["hdf5", "zarr"])
    def test_run_conversion_with_backend(self, setup_interface, tmp_path, backend):

        nwbfile_path = str(tmp_path / f"conversion_with_backend{backend}-{self.test_name}.nwb")

        metadata = self.interface.get_metadata()
        if "session_start_time" not in metadata["NWBFile"]:
            metadata["NWBFile"].update(session_start_time=datetime.now().astimezone())

        self.interface.run_conversion(
            nwbfile_path=nwbfile_path,
            overwrite=True,
            metadata=metadata,
            backend=backend,
            **self.conversion_options,
        )

        if backend == "zarr":
            with NWBZarrIO(path=nwbfile_path, mode="r") as io:
                io.read()

    @pytest.mark.parametrize("backend", ["hdf5", "zarr"])
    def test_configure_backend_for_equivalent_nwbfiles(self, setup_interface, tmp_path, backend):
        metadata = self.interface.get_metadata()
        if "session_start_time" not in metadata["NWBFile"]:
            metadata["NWBFile"].update(session_start_time=datetime.now().astimezone())

        nwbfile_1 = self.interface.create_nwbfile(metadata=metadata, **self.conversion_options)
        nwbfile_2 = self.interface.create_nwbfile(metadata=metadata, **self.conversion_options)

        backend_configuration = get_default_backend_configuration(nwbfile=nwbfile_1, backend=backend)
        configure_backend(nwbfile=nwbfile_2, backend_configuration=backend_configuration)

    def test_all_conversion_checks(self, setup_interface, tmp_path):
        interface, test_name = setup_interface

        # Create a unique test name and file path
        nwbfile_path = str(tmp_path / f"{self.__class__.__name__}_{self.test_name}.nwb")
        self.nwbfile_path = nwbfile_path

        # Now run the checks using the setup objects
        self.check_conversion_options_schema_valid()
        self.check_metadata()

        self.check_run_conversion_in_nwbconverter_with_backend(nwbfile_path=nwbfile_path, backend="hdf5")
        self.check_run_conversion_in_nwbconverter_with_backend_configuration(nwbfile_path=nwbfile_path, backend="hdf5")

        self.check_run_conversion_with_backend_configuration(nwbfile_path=nwbfile_path, backend="hdf5")

        self.check_read_nwb(nwbfile_path=nwbfile_path)

        # Any extra custom checks to run
        self.run_custom_checks()


class TemporalAlignmentMixin:
    """
    Generic class for testing temporal alignment methods.
    """

    data_interface_cls: Type[BaseDataInterface]
    interface_kwargs: dict
    save_directory: Path = Path(tempfile.mkdtemp())
    conversion_options: Optional[dict] = None
    maxDiff = None

    @pytest.fixture
    def setup_interface(self, request):

        self.test_name: str = ""
        self.interface = self.data_interface_cls(**self.interface_kwargs)
        return self.interface, self.test_name

    @pytest.fixture(scope="class", autouse=True)
    def setup_default_conversion_options(self, request):
        cls = request.cls
        cls.conversion_options = cls.conversion_options or dict()
        return cls.conversion_options

    def setUpFreshInterface(self):
        """Protocol for creating a fresh instance of the interface."""
        self.interface = self.data_interface_cls(**self.interface_kwargs)

    def check_interface_get_original_timestamps(self):
        """
        Just to ensure each interface can call .get_original_timestamps() without an error raising.

        Also, that it always returns non-empty.
        """
        self.setUpFreshInterface()
        original_timestamps = self.interface.get_original_timestamps()

        assert len(original_timestamps) != 0

    def check_interface_get_timestamps(self):
        """
        Just to ensure each interface can call .get_timestamps() without an error raising.

        Also, that it always returns non-empty.
        """
        self.setUpFreshInterface()
        timestamps = self.interface.get_timestamps()

        assert len(timestamps) != 0

    def check_interface_set_aligned_timestamps(self):
        """Ensure that internal mechanisms for the timestamps getter/setter work as expected."""
        self.setUpFreshInterface()
        unaligned_timestamps = self.interface.get_timestamps()

        random_number_generator = np.random.default_rng(seed=0)
        aligned_timestamps = (
            unaligned_timestamps + 1.23 + random_number_generator.random(size=unaligned_timestamps.shape)
        )
        self.interface.set_aligned_timestamps(aligned_timestamps=aligned_timestamps)

        retrieved_aligned_timestamps = self.interface.get_timestamps()
        assert_array_equal(x=retrieved_aligned_timestamps, y=aligned_timestamps)

    def check_shift_timestamps_by_start_time(self):
        """Ensure that internal mechanisms for shifting timestamps by a starting time work as expected."""
        self.setUpFreshInterface()
        unaligned_timestamps = self.interface.get_timestamps()

        aligned_starting_time = 1.23
        self.interface.set_aligned_starting_time(aligned_starting_time=aligned_starting_time)

        aligned_timestamps = self.interface.get_timestamps()
        expected_timestamps = unaligned_timestamps + aligned_starting_time
        assert_array_equal(x=aligned_timestamps, y=expected_timestamps)

    def check_interface_original_timestamps_inmutability(self):
        """Check aligning the timestamps for the interface does not change the value of .get_original_timestamps()."""
        self.setUpFreshInterface()
        pre_alignment_original_timestamps = self.interface.get_original_timestamps()

        aligned_timestamps = pre_alignment_original_timestamps + 1.23
        self.interface.set_aligned_timestamps(aligned_timestamps=aligned_timestamps)

        post_alignment_original_timestamps = self.interface.get_original_timestamps()
        assert_array_equal(x=post_alignment_original_timestamps, y=pre_alignment_original_timestamps)

    def check_nwbfile_temporal_alignment(self):
        """Check the temporally aligned timing information makes it into the NWB file."""
        pass  # TODO: will be easier to add when interface have 'add' methods separate from .run_conversion()

    def test_interface_alignment(self, setup_interface):

        interface, test_name = setup_interface

        self.check_interface_get_original_timestamps()
        self.check_interface_get_timestamps()
        self.check_interface_set_aligned_timestamps()
        self.check_shift_timestamps_by_start_time()
        self.check_interface_original_timestamps_inmutability()

        self.check_nwbfile_temporal_alignment()


class ImagingExtractorInterfaceTestMixin(DataInterfaceTestMixin, TemporalAlignmentMixin):
    data_interface_cls: Type[BaseImagingExtractorInterface]

    def check_read_nwb(self, nwbfile_path: str):
        from roiextractors import NwbImagingExtractor
        from roiextractors.testing import check_imaging_equal

        imaging = self.interface.imaging_extractor
        nwb_imaging = NwbImagingExtractor(file_path=nwbfile_path)

        exclude_channel_comparison = False
        if imaging.get_channel_names() is None:
            exclude_channel_comparison = True

        check_imaging_equal(imaging, nwb_imaging, exclude_channel_comparison)

    def check_nwbfile_temporal_alignment(self):
        nwbfile_path = str(
            self.save_directory
            / f"{self.data_interface_cls.__name__}_{self.test_name}_test_starting_time_alignment.nwb"
        )

        interface = self.data_interface_cls(**self.interface_kwargs)

        aligned_starting_time = 1.23
        interface.set_aligned_starting_time(aligned_starting_time=aligned_starting_time)

        metadata = interface.get_metadata()
        metadata["NWBFile"].update(session_start_time=datetime.now().astimezone())
        interface.run_conversion(nwbfile_path=nwbfile_path, overwrite=True, metadata=metadata)

        with NWBHDF5IO(path=nwbfile_path) as io:
            nwbfile = io.read()

            assert nwbfile.acquisition["TwoPhotonSeries"].starting_time == aligned_starting_time


class SegmentationExtractorInterfaceTestMixin(DataInterfaceTestMixin, TemporalAlignmentMixin):
    data_interface_cls: BaseSegmentationExtractorInterface

    def check_read(self, nwbfile_path: str):
        from roiextractors import NwbSegmentationExtractor
        from roiextractors.testing import check_segmentations_equal

        nwb_segmentation = NwbSegmentationExtractor(file_path=nwbfile_path)
        segmentation = self.interface.segmentation_extractor
        check_segmentations_equal(segmentation, nwb_segmentation)


class RecordingExtractorInterfaceTestMixin(DataInterfaceTestMixin, TemporalAlignmentMixin):
    """
    Generic class for testing any recording interface.
    """

    data_interface_cls: Type[BaseRecordingExtractorInterface]

    def check_read_nwb(self, nwbfile_path: str):
        from spikeinterface.extractors import NwbRecordingExtractor

        recording = self.interface.recording_extractor

        electrical_series_name = self.interface.get_metadata()["Ecephys"][self.interface.es_key]["name"]

        if recording.get_num_segments() == 1:
            # Spikeinterface behavior is to load the electrode table channel_name property as a channel_id
            self.nwb_recording = NwbRecordingExtractor(
                file_path=nwbfile_path,
                electrical_series_name=electrical_series_name,
                use_pynwb=True,
            )

            # Set channel_ids right for comparison
            # Neuroconv ALWAYS writes a string property `channel_name`` to the electrode table.
            # And the NwbRecordingExtractor always uses `channel_name` property as the channel_ids
            # `check_recordings_equal` compares ids so we need to rename the channels or the original recordings
            # So they match
            properties_in_the_recording = recording.get_property_keys()
            if "channel_name" in properties_in_the_recording:
                channel_name = recording.get_property("channel_name").astype("str", copy=False)
            else:
                channel_name = recording.get_channel_ids().astype("str", copy=False)

            recording = recording.rename_channels(new_channel_ids=channel_name)

            # Edge case that only occurs in testing, but should eventually be fixed nonetheless
            # The NwbRecordingExtractor on spikeinterface experiences an issue when duplicated channel_ids
            # are specified, which occurs during check_recordings_equal when there is only one channel
            if self.nwb_recording.get_channel_ids()[0] != self.nwb_recording.get_channel_ids()[-1]:
                check_recordings_equal(RX1=recording, RX2=self.nwb_recording, return_scaled=False)

                # This was added to test probe, we should just compare the probes
                for property_name in ["rel_x", "rel_y", "rel_z"]:
                    if (
                        property_name in properties_in_the_recording
                        or property_name in self.nwb_recording.get_property_keys()
                    ):
                        assert_array_equal(
                            recording.get_property(property_name), self.nwb_recording.get_property(property_name)
                        )
                if recording.has_scaled_traces() and self.nwb_recording.has_scaled_traces():
                    check_recordings_equal(RX1=recording, RX2=self.nwb_recording, return_scaled=True)

            # Compare channel groups
            # Neuroconv ALWAYS writes a string property `group_name` to the electrode table.
            # The NwbRecordingExtractor takes the `group_name` from the electrode table and sets it `group` property
            if "group_name" in properties_in_the_recording:
                group_name_array = recording.get_property("group_name").astype("str", copy=False)
            elif "group" in properties_in_the_recording:
                group_name_array = recording.get_property("group").astype("str", copy=False)
            else:
                default_group_name = "ElectrodeGroup"
                group_name_array = np.full(channel_name.size, fill_value=default_group_name)

            group_names_in_nwb = self.nwb_recording.get_property("group")
            np.testing.assert_array_equal(group_name_array, group_names_in_nwb)

    def check_interface_set_aligned_timestamps(self):
        self.setUpFreshInterface()

        random_number_generator = np.random.default_rng(seed=0)
        if self.interface._number_of_segments == 1:
            unaligned_timestamps = self.interface.get_timestamps()

            aligned_timestamps = (
                unaligned_timestamps + 1.23 + random_number_generator.random(size=unaligned_timestamps.shape)
            )
            self.interface.set_aligned_timestamps(aligned_timestamps=aligned_timestamps)

            retrieved_aligned_timestamps = self.interface.get_timestamps()
            assert_array_equal(x=retrieved_aligned_timestamps, y=aligned_timestamps)
        else:
            with pytest.raises(
                AssertionError,
                match="This recording has multiple segments; please use 'align_segment_timestamps' instead.",
            ):
                all_unaligned_timestamps = self.interface.get_timestamps()

                all_aligned_segment_timestamps = [
                    unaligned_timestamps + 1.23 + random_number_generator.random(size=unaligned_timestamps.shape)
                    for unaligned_timestamps in all_unaligned_timestamps
                ]
                self.interface.set_aligned_timestamps(aligned_timestamps=all_aligned_segment_timestamps)

    def check_interface_set_aligned_segment_timestamps(self):
        self.setUpFreshInterface()

        random_number_generator = np.random.default_rng(seed=0)
        if self.interface._number_of_segments == 1:
            unaligned_timestamps = self.interface.get_timestamps()

            all_aligned_segment_timestamps = [
                unaligned_timestamps + 1.23 + random_number_generator.random(size=unaligned_timestamps.shape)
            ]
            self.interface.set_aligned_segment_timestamps(aligned_segment_timestamps=all_aligned_segment_timestamps)

            retrieved_aligned_timestamps = self.interface.get_timestamps()
            assert_array_equal(x=retrieved_aligned_timestamps, y=all_aligned_segment_timestamps[0])
        else:
            all_unaligned_timestamps = self.interface.get_timestamps()
            all_aligned_segment_timestamps = [
                unaligned_timestamps + 1.23 + random_number_generator.random(size=unaligned_timestamps.shape)
                for unaligned_timestamps in all_unaligned_timestamps
            ]
            self.interface.set_aligned_segment_timestamps(aligned_segment_timestamps=all_aligned_segment_timestamps)

            all_retrieved_aligned_timestamps = self.interface.get_timestamps()
            for retrieved_aligned_timestamps, aligned_segment_timestamps in zip(
                all_retrieved_aligned_timestamps, all_aligned_segment_timestamps
            ):
                assert_array_equal(x=retrieved_aligned_timestamps, y=aligned_segment_timestamps)

    def check_shift_timestamps_by_start_time(self):
        self.setUpFreshInterface()
        all_unaligned_timestamps = self.interface.get_timestamps()

        aligned_starting_time = 1.23
        self.interface.set_aligned_starting_time(aligned_starting_time=aligned_starting_time)

        if self.interface._number_of_segments == 1:
            retrieved_aligned_timestamps = self.interface.get_timestamps()
            expected_timestamps = all_unaligned_timestamps + aligned_starting_time
            assert_array_equal(x=retrieved_aligned_timestamps, y=expected_timestamps)
        else:
            all_retrieved_aligned_timestamps = self.interface.get_timestamps()
            all_expected_timestamps = [
                unaligned_timestamps + aligned_starting_time for unaligned_timestamps in all_unaligned_timestamps
            ]
            for retrieved_aligned_timestamps, expected_timestamps in zip(
                all_retrieved_aligned_timestamps, all_expected_timestamps
            ):
                assert_array_equal(x=retrieved_aligned_timestamps, y=expected_timestamps)

    def check_shift_segment_timestamps_by_starting_times(self):
        self.setUpFreshInterface()

        aligned_segment_starting_times = list(np.arange(float(self.interface._number_of_segments)) + 1.23)
        if self.interface._number_of_segments == 1:
            unaligned_timestamps = self.interface.get_timestamps()

            self.interface.set_aligned_segment_starting_times(
                aligned_segment_starting_times=aligned_segment_starting_times
            )

            retrieved_aligned_timestamps = self.interface.get_timestamps()
            expected_aligned_timestamps = unaligned_timestamps + aligned_segment_starting_times[0]
            assert_array_equal(x=retrieved_aligned_timestamps, y=expected_aligned_timestamps)
        else:
            all_unaligned_timestamps = self.interface.get_timestamps()

            self.interface.set_aligned_segment_starting_times(
                aligned_segment_starting_times=aligned_segment_starting_times
            )

            all_retrieved_aligned_timestamps = self.interface.get_timestamps()
            all_expected_aligned_timestamps = [
                timestamps + segment_starting_time
                for timestamps, segment_starting_time in zip(all_unaligned_timestamps, aligned_segment_starting_times)
            ]
            for retrieved_aligned_timestamps, expected_aligned_timestamps in zip(
                all_retrieved_aligned_timestamps, all_expected_aligned_timestamps
            ):
                assert_array_equal(x=retrieved_aligned_timestamps, y=expected_aligned_timestamps)

    def check_interface_original_timestamps_inmutability(self):
        """Check aligning the timestamps for the interface does not change the value of .get_original_timestamps()."""
        self.setUpFreshInterface()

        if self.interface._number_of_segments == 1:
            pre_alignment_original_timestamps = self.interface.get_original_timestamps()

            aligned_timestamps = pre_alignment_original_timestamps + 1.23
            self.interface.set_aligned_timestamps(aligned_timestamps=aligned_timestamps)

            post_alignment_original_timestamps = self.interface.get_original_timestamps()
            assert_array_equal(x=post_alignment_original_timestamps, y=pre_alignment_original_timestamps)
        else:
            with pytest.raises(
                AssertionError,
                match="This recording has multiple segments; please use 'align_segment_timestamps' instead.",
            ):
                all_pre_alignment_timestamps = self.interface.get_original_timestamps()

                all_aligned_timestamps = [
                    unaligned_timestamps + 1.23 for unaligned_timestamps in all_pre_alignment_timestamps
                ]
                self.interface.set_aligned_timestamps(aligned_timestamps=all_aligned_timestamps)

    def test_interface_alignment(self, setup_interface):

        interface, test_name = setup_interface

        self.check_interface_get_original_timestamps()
        self.check_interface_get_timestamps()
        self.check_interface_set_aligned_timestamps()
        self.check_shift_timestamps_by_start_time()
        self.check_interface_original_timestamps_inmutability()

        self.check_interface_set_aligned_segment_timestamps()
        self.check_shift_timestamps_by_start_time()
        self.check_shift_segment_timestamps_by_starting_times()

        self.check_nwbfile_temporal_alignment()


class SortingExtractorInterfaceTestMixin(DataInterfaceTestMixin, TemporalAlignmentMixin):
    data_interface_cls: Type[BaseSortingExtractorInterface]
    associated_recording_cls: Optional[Type[BaseRecordingExtractorInterface]] = None
    associated_recording_kwargs: Optional[dict] = None

    def setUpFreshInterface(self):
        self.interface = self.data_interface_cls(**self.interface_kwargs)

        recording_interface = self.associated_recording_cls(**self.associated_recording_kwargs)
        self.interface.register_recording(recording_interface=recording_interface)

    def check_read_nwb(self, nwbfile_path: str):
        from spikeinterface.extractors import NwbSortingExtractor

        sorting = self.interface.sorting_extractor
        sf = sorting.get_sampling_frequency()
        if sf is None:  # need to set dummy sampling frequency since no associated acquisition in file
            sorting.set_sampling_frequency(30_000.0)

        # NWBSortingExtractor on spikeinterface does not yet support loading data written from multiple segment.
        if sorting.get_num_segments() == 1:
            # TODO after 0.100 release remove this if
            signature = inspect.signature(NwbSortingExtractor)
            if "t_start" in signature.parameters:
                nwb_sorting = NwbSortingExtractor(file_path=nwbfile_path, sampling_frequency=sf, t_start=0.0)
            else:
                nwb_sorting = NwbSortingExtractor(file_path=nwbfile_path, sampling_frequency=sf)
            # In the NWBSortingExtractor, since unit_names could be not unique,
            # table "ids" are loaded as unit_ids. Here we rename the original sorting accordingly
            if "unit_name" in sorting.get_property_keys():
                renamed_unit_ids = sorting.get_property("unit_name")
                # sorting_renamed = sorting.rename_units(new_unit_ids=renamed_unit_ids)  #TODO after 0.100 release use this
                sorting_renamed = sorting.select_units(unit_ids=sorting.unit_ids, renamed_unit_ids=renamed_unit_ids)

            else:
                nwb_has_ids_as_strings = all(isinstance(id, str) for id in nwb_sorting.unit_ids)
                if nwb_has_ids_as_strings:
                    renamed_unit_ids = sorting.get_unit_ids()
                    renamed_unit_ids = [str(id) for id in renamed_unit_ids]
                else:
                    renamed_unit_ids = np.arange(len(sorting.unit_ids))

                # sorting_renamed = sorting.rename_units(new_unit_ids=sorting.unit_ids) #TODO after 0.100 release use this
                sorting_renamed = sorting.select_units(unit_ids=sorting.unit_ids, renamed_unit_ids=renamed_unit_ids)
            check_sortings_equal(SX1=sorting_renamed, SX2=nwb_sorting)

    def check_interface_set_aligned_segment_timestamps(self):
        self.setUpFreshInterface()

        if self.interface.sorting_extractor.has_recording():
            random_number_generator = np.random.default_rng(seed=0)
            if self.interface._number_of_segments == 1:
                unaligned_timestamps = self.interface.get_timestamps()

                all_aligned_segment_timestamps = [
                    unaligned_timestamps + 1.23 + random_number_generator.random(size=unaligned_timestamps.shape)
                ]
                self.interface.set_aligned_segment_timestamps(aligned_segment_timestamps=all_aligned_segment_timestamps)

                retrieved_aligned_timestamps = self.interface.get_timestamps()
                assert_array_equal(x=retrieved_aligned_timestamps, y=all_aligned_segment_timestamps[0])
            else:
                all_unaligned_timestamps = self.interface.get_timestamps()
                all_aligned_segment_timestamps = [
                    unaligned_timestamps + 1.23 + random_number_generator.random(size=unaligned_timestamps.shape)
                    for unaligned_timestamps in all_unaligned_timestamps
                ]
                self.interface.set_aligned_segment_timestamps(aligned_segment_timestamps=all_aligned_segment_timestamps)

                all_retrieved_aligned_timestamps = self.interface.get_timestamps()
                for retrieved_aligned_timestamps, aligned_segment_timestamps in zip(
                    all_retrieved_aligned_timestamps, all_aligned_segment_timestamps
                ):
                    assert_array_equal(x=retrieved_aligned_timestamps, y=aligned_segment_timestamps)

    def check_shift_segment_timestamps_by_starting_times(self):
        self.setUpFreshInterface()

        aligned_segment_starting_times = list(np.arange(float(self.interface._number_of_segments)) + 1.23)
        if self.interface._number_of_segments == 1:
            unaligned_timestamps = self.interface.get_timestamps()

            self.interface.set_aligned_segment_starting_times(
                aligned_segment_starting_times=aligned_segment_starting_times
            )

            retrieved_aligned_timestamps = self.interface.get_timestamps()
            expected_aligned_timestamps = unaligned_timestamps + aligned_segment_starting_times[0]
            assert_array_equal(x=retrieved_aligned_timestamps, y=expected_aligned_timestamps)
        else:
            all_unaligned_timestamps = self.interface.get_timestamps()

            self.interface.set_aligned_segment_starting_times(
                aligned_segment_starting_times=aligned_segment_starting_times
            )

            all_retrieved_aligned_timestamps = self.interface.get_timestamps()
            all_expected_aligned_timestamps = [
                segment_timestamps + segment_starting_time
                for segment_timestamps, segment_starting_time in zip(
                    all_unaligned_timestamps, aligned_segment_starting_times
                )
            ]
            for retrieved_aligned_timestamps, expected_aligned_timestamps in zip(
                all_retrieved_aligned_timestamps, all_expected_aligned_timestamps
            ):
                assert_array_equal(x=retrieved_aligned_timestamps, y=expected_aligned_timestamps)

    def test_all_conversion_checks(self, setup_interface, tmp_path):
        # The fixture `setup_interface` sets up the necessary objects
        interface, test_name = setup_interface

        # Create a unique test name and file path
        nwbfile_path = str(tmp_path / f"{self.__class__.__name__}_{self.test_name}.nwb")

        # Now run the checks using the setup objects
        self.check_conversion_options_schema_valid()
        self.check_metadata()

        self.check_run_conversion_in_nwbconverter_with_backend(nwbfile_path=nwbfile_path, backend="hdf5")
        self.check_run_conversion_in_nwbconverter_with_backend_configuration(nwbfile_path=nwbfile_path, backend="hdf5")

        self.check_run_conversion_with_backend_configuration(nwbfile_path=nwbfile_path, backend="hdf5")

        self.check_read_nwb(nwbfile_path=nwbfile_path)

        # Any extra custom checks to run
        self.run_custom_checks()

    def test_interface_alignment(self, setup_interface):

        # TODO sorting can have times without associated recordings, test this later
        if self.associated_recording_cls is None:
            return None

        # Skip get_original_timestamps() checks since unsupported
        self.check_interface_get_timestamps()
        self.check_interface_set_aligned_timestamps()
        self.check_interface_set_aligned_segment_timestamps()
        self.check_shift_timestamps_by_start_time()
        self.check_shift_segment_timestamps_by_starting_times()

        self.check_nwbfile_temporal_alignment()


class AudioInterfaceTestMixin(DataInterfaceTestMixin, TemporalAlignmentMixin):
    """
    A mixin for testing Audio interfaces.
    """

    # Currently asserted in the downstream testing suite; could be refactored in future PR
    def check_read_nwb(self, nwbfile_path: str):
        pass

    # Currently asserted in the downstream testing suite
    def test_interface_alignment(self):
        pass


class DeepLabCutInterfaceMixin(DataInterfaceTestMixin, TemporalAlignmentMixin):
    """
    A mixin for testing DeepLabCut interfaces.
    """

    def check_interface_get_original_timestamps(self):
        pass  # TODO in separate PR

    def check_interface_get_timestamps(self):
        pass  # TODO in separate PR

    def check_interface_set_aligned_timestamps(self):
        pass  # TODO in separate PR

    def check_shift_timestamps_by_start_time(self):
        pass  # TODO in separate PR

    def check_interface_original_timestamps_inmutability(self):
        pass  # TODO in separate PR

    def check_nwbfile_temporal_alignment(self):
        pass  # TODO in separate PR


class VideoInterfaceMixin(DataInterfaceTestMixin, TemporalAlignmentMixin):
    """
    A mixin for testing Video interfaces.
    """

    def check_read_nwb(self, nwbfile_path: str):
        with NWBHDF5IO(path=nwbfile_path, mode="r", load_namespaces=True) as io:
            nwbfile = io.read()
            video_type = Path(self.interface_kwargs["file_paths"][0]).suffix[1:]
            assert f"Video video_{video_type}" in nwbfile.acquisition

    def check_interface_set_aligned_timestamps(self):
        all_unaligned_timestamps = self.interface.get_original_timestamps()

        random_number_generator = np.random.default_rng(seed=0)
        aligned_timestamps = [
            unaligned_timestamps + 1.23 + random_number_generator.random(size=unaligned_timestamps.shape)
            for unaligned_timestamps in all_unaligned_timestamps
        ]
        self.interface.set_aligned_timestamps(aligned_timestamps=aligned_timestamps)

        retrieved_aligned_timestamps = self.interface.get_timestamps()
        assert_array_equal(x=retrieved_aligned_timestamps, y=aligned_timestamps)

    def check_shift_timestamps_by_start_time(self):
        self.setUpFreshInterface()

        aligned_starting_time = 1.23
        self.interface.set_aligned_timestamps(aligned_timestamps=self.interface.get_original_timestamps())
        self.interface.set_aligned_starting_time(aligned_starting_time=aligned_starting_time)
        all_aligned_timestamps = self.interface.get_timestamps()

        unaligned_timestamps = self.interface.get_original_timestamps()
        all_expected_timestamps = [timestamps + aligned_starting_time for timestamps in unaligned_timestamps]
        [
            assert_array_equal(x=aligned_timestamps, y=expected_timestamps)
            for aligned_timestamps, expected_timestamps in zip(all_aligned_timestamps, all_expected_timestamps)
        ]

    def check_set_aligned_segment_starting_times(self):
        self.setUpFreshInterface()

        aligned_segment_starting_times = [
            1.23 * file_path_index for file_path_index in range(len(self.interface_kwargs))
        ]
        self.interface.set_aligned_segment_starting_times(aligned_segment_starting_times=aligned_segment_starting_times)
        all_aligned_timestamps = self.interface.get_timestamps()

        unaligned_timestamps = self.interface.get_original_timestamps()
        all_expected_timestamps = [
            timestamps + segment_starting_time
            for timestamps, segment_starting_time in zip(unaligned_timestamps, aligned_segment_starting_times)
        ]
        for aligned_timestamps, expected_timestamps in zip(all_aligned_timestamps, all_expected_timestamps):
            assert_array_equal(x=aligned_timestamps, y=expected_timestamps)

    def check_interface_original_timestamps_inmutability(self):
        self.setUpFreshInterface()

        all_pre_alignment_original_timestamps = self.interface.get_original_timestamps()

        all_aligned_timestamps = [
            pre_alignment_original_timestamps + 1.23
            for pre_alignment_original_timestamps in all_pre_alignment_original_timestamps
        ]
        self.interface.set_aligned_timestamps(aligned_timestamps=all_aligned_timestamps)

        all_post_alignment_original_timestamps = self.interface.get_original_timestamps()
        for post_alignment_original_timestamps, pre_alignment_original_timestamps in zip(
            all_post_alignment_original_timestamps, all_pre_alignment_original_timestamps
        ):
            assert_array_equal(x=post_alignment_original_timestamps, y=pre_alignment_original_timestamps)


class MedPCInterfaceMixin(DataInterfaceTestMixin, TemporalAlignmentMixin):
    """
    A mixin for testing MedPC interfaces.
    """

    def test_metadata_schema_valid(self):
        pass

    def test_run_conversion_with_backend(self):
        pass

    def test_no_metadata_mutation(self):
        pass

    def test_configure_backend_for_equivalent_nwbfiles(self):
        pass

    def check_metadata_schema_valid(self):
        schema = self.interface.get_metadata_schema()
        Draft7Validator.check_schema(schema=schema)

    def check_metadata(self):
        schema = self.interface.get_metadata_schema()
        metadata = self.interface.get_metadata()
        if "session_start_time" not in metadata["NWBFile"]:
            metadata["NWBFile"].update(session_start_time=datetime.now().astimezone())
        # handle json encoding of datetimes and other tricky types
        metadata_for_validation = json.loads(json.dumps(metadata, cls=_NWBMetaDataEncoder))
        validate(metadata_for_validation, schema)
        self.check_extracted_metadata(metadata)

    def check_no_metadata_mutation(self, metadata: dict):
        """Ensure the metadata object was not altered by `add_to_nwbfile` method."""

        metadata_in = deepcopy(metadata)

        nwbfile = mock_NWBFile()
        self.interface.add_to_nwbfile(nwbfile=nwbfile, metadata=metadata, **self.conversion_options)

        assert metadata == metadata_in

    def check_run_conversion_with_backend(
        self, nwbfile_path: str, metadata: dict, backend: Literal["hdf5", "zarr"] = "hdf5"
    ):
        self.interface.run_conversion(
            nwbfile_path=nwbfile_path,
            overwrite=True,
            metadata=metadata,
            backend=backend,
            **self.conversion_options,
        )

    def check_configure_backend_for_equivalent_nwbfiles(
        self, metadata: dict, backend: Literal["hdf5", "zarr"] = "hdf5"
    ):
        nwbfile_1 = self.interface.create_nwbfile(metadata=metadata, **self.conversion_options)
        nwbfile_2 = self.interface.create_nwbfile(metadata=metadata, **self.conversion_options)

        backend_configuration = get_default_backend_configuration(nwbfile=nwbfile_1, backend=backend)
        configure_backend(nwbfile=nwbfile_2, backend_configuration=backend_configuration)

    def check_run_conversion_with_backend_configuration(
        self, nwbfile_path: str, metadata: dict, backend: Literal["hdf5", "zarr"] = "hdf5"
    ):
        nwbfile = self.interface.create_nwbfile(metadata=metadata, **self.conversion_options)
        backend_configuration = self.interface.get_default_backend_configuration(nwbfile=nwbfile, backend=backend)
        self.interface.run_conversion(
            nwbfile_path=nwbfile_path,
            nwbfile=nwbfile,
            overwrite=True,
            metadata=metadata,
            backend_configuration=backend_configuration,
            **self.conversion_options,
        )

    def check_run_conversion_in_nwbconverter_with_backend(
        self, nwbfile_path: str, metadata: dict, backend: Literal["hdf5", "zarr"] = "hdf5"
    ):
        class TestNWBConverter(NWBConverter):
            data_interface_classes = dict(Test=type(self.interface))

        test_kwargs = self.test_kwargs[0] if isinstance(self.test_kwargs, list) else self.test_kwargs
        source_data = dict(Test=test_kwargs)
        converter = TestNWBConverter(source_data=source_data)

        conversion_options = dict(Test=self.conversion_options)
        converter.run_conversion(
            nwbfile_path=nwbfile_path,
            overwrite=True,
            metadata=metadata,
            backend=backend,
            conversion_options=conversion_options,
        )

    def check_run_conversion_in_nwbconverter_with_backend_configuration(
        self, nwbfile_path: str, metadata: dict, backend: Union["hdf5", "zarr"] = "hdf5"
    ):
        class TestNWBConverter(NWBConverter):
            data_interface_classes = dict(Test=type(self.interface))

        test_kwargs = self.test_kwargs[0] if isinstance(self.test_kwargs, list) else self.test_kwargs
        source_data = dict(Test=test_kwargs)
        converter = TestNWBConverter(source_data=source_data)

        conversion_options = dict(Test=self.conversion_options)

        nwbfile = converter.create_nwbfile(metadata=metadata, conversion_options=conversion_options)
        backend_configuration = converter.get_default_backend_configuration(nwbfile=nwbfile, backend=backend)
        converter.run_conversion(
            nwbfile_path=nwbfile_path,
            nwbfile=nwbfile,
            overwrite=True,
            metadata=metadata,
            backend_configuration=backend_configuration,
            conversion_options=conversion_options,
        )

    def test_all_conversion_checks(self, metadata: dict):
        interface_kwargs = self.interface_kwargs
        if isinstance(interface_kwargs, dict):
            interface_kwargs = [interface_kwargs]
        for num, kwargs in enumerate(interface_kwargs):
            with self.subTest(str(num)):
                self.case = num
                self.test_kwargs = kwargs
                self.interface = self.data_interface_cls(**self.test_kwargs)

                self.check_metadata_schema_valid()
                self.check_conversion_options_schema_valid()
                self.check_metadata()
                self.nwbfile_path = str(self.save_directory / f"{self.__class__.__name__}_{num}.nwb")

                self.check_no_metadata_mutation(metadata=metadata)

                self.check_configure_backend_for_equivalent_nwbfiles(metadata=metadata)

                self.check_run_conversion_in_nwbconverter_with_backend(
                    nwbfile_path=self.nwbfile_path, metadata=metadata, backend="hdf5"
                )
                self.check_run_conversion_in_nwbconverter_with_backend_configuration(
                    nwbfile_path=self.nwbfile_path, metadata=metadata, backend="hdf5"
                )

                self.check_run_conversion_with_backend(
                    nwbfile_path=self.nwbfile_path, metadata=metadata, backend="hdf5"
                )
                self.check_run_conversion_with_backend_configuration(
                    nwbfile_path=self.nwbfile_path, metadata=metadata, backend="hdf5"
                )

                self.check_read_nwb(nwbfile_path=self.nwbfile_path)

                # TODO: enable when all H5DataIO prewraps are gone
                # self.nwbfile_path = str(self.save_directory / f"{self.__class__.__name__}_{num}.nwb.zarr")
                # self.check_run_conversion(nwbfile_path=self.nwbfile_path, backend="zarr")
                # self.check_run_conversion_custom_backend(nwbfile_path=self.nwbfile_path, backend="zarr")
                # self.check_basic_zarr_read(nwbfile_path=self.nwbfile_path)

                # Any extra custom checks to run
                self.run_custom_checks()

    def check_interface_get_original_timestamps(self, medpc_name_to_info_dict: dict):
        """
        Just to ensure each interface can call .get_original_timestamps() without an error raising.

        Also, that it always returns non-empty.
        """
        self.setUpFreshInterface()
        original_timestamps_dict = self.interface.get_original_timestamps(
            medpc_name_to_info_dict=medpc_name_to_info_dict
        )
        for name in self.interface.source_data["aligned_timestamp_names"]:
            original_timestamps = original_timestamps_dict[name]
            assert len(original_timestamps) != 0, f"Timestamps for {name} are empty."

    def check_interface_get_timestamps(self):
        """
        Just to ensure each interface can call .get_timestamps() without an error raising.

        Also, that it always returns non-empty.
        """
        self.setUpFreshInterface()
        timestamps_dict = self.interface.get_timestamps()
        for timestamps in timestamps_dict.values():
            assert len(timestamps) != 0

    def check_interface_set_aligned_timestamps(self, medpc_name_to_info_dict: dict):
        """Ensure that internal mechanisms for the timestamps getter/setter work as expected."""
        self.setUpFreshInterface()
        unaligned_timestamps_dict = self.interface.get_original_timestamps(
            medpc_name_to_info_dict=medpc_name_to_info_dict
        )

        random_number_generator = np.random.default_rng(seed=0)
        aligned_timestamps_dict = {}
        for name, unaligned_timestamps in unaligned_timestamps_dict.items():
            aligned_timestamps = (
                unaligned_timestamps + 1.23 + random_number_generator.random(size=unaligned_timestamps.shape)
            )
            aligned_timestamps_dict[name] = aligned_timestamps
        self.interface.set_aligned_timestamps(aligned_timestamps_dict=aligned_timestamps_dict)

        retrieved_aligned_timestamps = self.interface.get_timestamps()
        for name, aligned_timestamps in aligned_timestamps_dict.items():
            assert_array_equal(retrieved_aligned_timestamps[name], aligned_timestamps)

    def check_shift_timestamps_by_start_time(self, medpc_name_to_info_dict: dict):
        """Ensure that internal mechanisms for shifting timestamps by a starting time work as expected."""
        self.setUpFreshInterface()
        unaligned_timestamps_dict = self.interface.get_original_timestamps(
            medpc_name_to_info_dict=medpc_name_to_info_dict
        )

        aligned_starting_time = 1.23
        self.interface.set_aligned_starting_time(
            aligned_starting_time=aligned_starting_time,
            medpc_name_to_info_dict=medpc_name_to_info_dict,
        )

        aligned_timestamps = self.interface.get_timestamps()
        expected_timestamps_dict = {
            name: unaligned_timestamps + aligned_starting_time
            for name, unaligned_timestamps in unaligned_timestamps_dict.items()
        }
        for name, expected_timestamps in expected_timestamps_dict.items():
            assert_array_equal(aligned_timestamps[name], expected_timestamps)

    def check_interface_original_timestamps_inmutability(self, medpc_name_to_info_dict: dict):
        """Check aligning the timestamps for the interface does not change the value of .get_original_timestamps()."""
        self.setUpFreshInterface()
        pre_alignment_original_timestamps_dict = self.interface.get_original_timestamps(
            medpc_name_to_info_dict=medpc_name_to_info_dict
        )

        aligned_timestamps_dict = {
            name: pre_alignment_og_timestamps + 1.23
            for name, pre_alignment_og_timestamps in pre_alignment_original_timestamps_dict.items()
        }
        self.interface.set_aligned_timestamps(aligned_timestamps_dict=aligned_timestamps_dict)

        post_alignment_original_timestamps_dict = self.interface.get_original_timestamps(
            medpc_name_to_info_dict=medpc_name_to_info_dict
        )
        for name, post_alignment_original_timestamps_dict in post_alignment_original_timestamps_dict.items():
            assert_array_equal(post_alignment_original_timestamps_dict, pre_alignment_original_timestamps_dict[name])

    def test_interface_alignment(self, medpc_name_to_info_dict: dict):
        interface_kwargs = self.interface_kwargs
        if isinstance(interface_kwargs, dict):
            interface_kwargs = [interface_kwargs]
        for num, kwargs in enumerate(interface_kwargs):
            with self.subTest(str(num)):
                self.case = num
                self.test_kwargs = kwargs

                self.check_interface_get_original_timestamps(medpc_name_to_info_dict=medpc_name_to_info_dict)
                self.check_interface_get_timestamps()
                self.check_interface_set_aligned_timestamps(medpc_name_to_info_dict=medpc_name_to_info_dict)
                self.check_shift_timestamps_by_start_time(medpc_name_to_info_dict=medpc_name_to_info_dict)
                self.check_interface_original_timestamps_inmutability(medpc_name_to_info_dict=medpc_name_to_info_dict)

                self.check_nwbfile_temporal_alignment()


class MiniscopeImagingInterfaceMixin(DataInterfaceTestMixin, TemporalAlignmentMixin):
    """
    A mixin for testing Miniscope Imaging interfaces.
    """

    def check_read_nwb(self, nwbfile_path: str):
        from ndx_miniscope import Miniscope

        with NWBHDF5IO(nwbfile_path, "r") as io:
            nwbfile = io.read()

            assert self.device_name in nwbfile.devices
            device = nwbfile.devices[self.device_name]
            assert isinstance(device, Miniscope)
            imaging_plane = nwbfile.imaging_planes[self.imaging_plane_name]
            assert imaging_plane.device.name == self.device_name

            # Check OnePhotonSeries
            assert self.photon_series_name in nwbfile.acquisition
            one_photon_series = nwbfile.acquisition[self.photon_series_name]
            assert one_photon_series.unit == "px"
            assert one_photon_series.data.shape == (15, 752, 480)
            assert one_photon_series.data.dtype == np.uint8
            assert one_photon_series.rate is None
            assert one_photon_series.starting_frame is None
            assert one_photon_series.timestamps.shape == (15,)

            imaging_extractor = self.interface.imaging_extractor
            times_from_extractor = imaging_extractor._times
            assert_array_equal(one_photon_series.timestamps, times_from_extractor)


class ScanImageSinglePlaneImagingInterfaceMixin(DataInterfaceTestMixin, TemporalAlignmentMixin):
    """
    A mixing for testing ScanImage Single Plane Imaging interfaces.
    """

    def check_read_nwb(self, nwbfile_path: str):
        with NWBHDF5IO(nwbfile_path, "r") as io:
            nwbfile = io.read()

            assert self.imaging_plane_name in nwbfile.imaging_planes
            assert self.photon_series_name in nwbfile.acquisition
            photon_series_suffix = self.photon_series_name.replace("TwoPhotonSeries", "")
            assert self.interface.two_photon_series_name_suffix == photon_series_suffix
            two_photon_series = nwbfile.acquisition[self.photon_series_name]
            assert two_photon_series.data.shape == self.expected_two_photon_series_data_shape
            assert two_photon_series.unit == "n.a."
            assert two_photon_series.data.dtype == np.int16
            assert two_photon_series.rate is None
            assert two_photon_series.starting_time is None

            imaging_extractor = self.interface.imaging_extractor
            times_from_extractor = imaging_extractor._times
            assert_array_equal(two_photon_series.timestamps[:], times_from_extractor)

            data_from_extractor = imaging_extractor.get_video()
            assert_array_equal(two_photon_series.data[:], data_from_extractor.transpose(0, 2, 1))

            assert two_photon_series.description == json.dumps(self.interface.image_metadata)

            optical_channels = nwbfile.imaging_planes[self.imaging_plane_name].optical_channel
            optical_channel_names = [channel.name for channel in optical_channels]
            assert self.interface_kwargs["channel_name"] in optical_channel_names
            assert len(optical_channels) == 1


class ScanImageMultiPlaneImagingInterfaceMixin(DataInterfaceTestMixin, TemporalAlignmentMixin):
    """
    A mixin for testing ScanImage MultiPlane Imaging interfaces.
    """

    def check_read_nwb(self, nwbfile_path: str):
        with NWBHDF5IO(nwbfile_path, "r") as io:
            nwbfile = io.read()

            assert self.imaging_plane_name in nwbfile.imaging_planes
            assert self.photon_series_name in nwbfile.acquisition
            photon_series_suffix = self.photon_series_name.replace("TwoPhotonSeries", "")
            assert self.interface.two_photon_series_name_suffix == photon_series_suffix
            two_photon_series = nwbfile.acquisition[self.photon_series_name]
            assert two_photon_series.data.shape == self.expected_two_photon_series_data_shape
            assert two_photon_series.unit == "n.a."
            assert two_photon_series.data.dtype == np.int16

            assert two_photon_series.rate == self.expected_rate
            assert two_photon_series.starting_time == self.expected_starting_time
            assert two_photon_series.timestamps is None

            imaging_extractor = self.interface.imaging_extractor
            data_from_extractor = imaging_extractor.get_video()
            assert_array_equal(two_photon_series.data[:], data_from_extractor.transpose(0, 2, 1, 3))

            assert two_photon_series.description == json.dumps(imaging_extractor._imaging_extractors[0].metadata)

            optical_channels = nwbfile.imaging_planes[self.imaging_plane_name].optical_channel
            optical_channel_names = [channel.name for channel in optical_channels]
            assert self.interface_kwargs["channel_name"] in optical_channel_names
            assert len(optical_channels) == 1


class TDTFiberPhotometryInterfaceMixin(DataInterfaceTestMixin, TemporalAlignmentMixin):
    """Mixin for testing TDT Fiber Photometry interfaces."""

    def test_metadata_schema_valid(self):
        pass

    def test_no_metadata_mutation(self):
        pass

    def test_run_conversion_with_backend(self):
        pass

    def test_no_metadata_mutation(self):
        pass

    def test_configure_backend_for_equivalent_nwbfiles(self):
        pass

    def check_metadata_schema_valid(self):
        schema = self.interface.get_metadata_schema()
        Draft7Validator.check_schema(schema=schema)

    def check_no_metadata_mutation(self, metadata: dict):
        """Ensure the metadata object was not altered by `add_to_nwbfile` method."""

        metadata_in = deepcopy(metadata)

        nwbfile = mock_NWBFile()
        self.interface.add_to_nwbfile(nwbfile=nwbfile, metadata=metadata, **self.conversion_options)

        assert metadata == metadata_in

    def check_run_conversion_with_backend(
        self, nwbfile_path: str, metadata: dict, backend: Literal["hdf5", "zarr"] = "hdf5"
    ):
        self.interface.run_conversion(
            nwbfile_path=nwbfile_path,
            overwrite=True,
            metadata=metadata,
            backend=backend,
            **self.conversion_options,
        )

    def check_configure_backend_for_equivalent_nwbfiles(
        self, metadata: dict, backend: Literal["hdf5", "zarr"] = "hdf5"
    ):
        nwbfile_1 = self.interface.create_nwbfile(metadata=metadata, **self.conversion_options)
        nwbfile_2 = self.interface.create_nwbfile(metadata=metadata, **self.conversion_options)

        backend_configuration = get_default_backend_configuration(nwbfile=nwbfile_1, backend=backend)
        configure_backend(nwbfile=nwbfile_2, backend_configuration=backend_configuration)

    def check_run_conversion_with_backend_configuration(
        self, nwbfile_path: str, metadata: dict, backend: Literal["hdf5", "zarr"] = "hdf5"
    ):
        nwbfile = self.interface.create_nwbfile(metadata=metadata, **self.conversion_options)
        backend_configuration = self.interface.get_default_backend_configuration(nwbfile=nwbfile, backend=backend)
        self.interface.run_conversion(
            nwbfile_path=nwbfile_path,
            nwbfile=nwbfile,
            overwrite=True,
            backend_configuration=backend_configuration,
            **self.conversion_options,
        )

    def check_run_conversion_in_nwbconverter_with_backend(
        self, nwbfile_path: str, metadata: dict, backend: Literal["hdf5", "zarr"] = "hdf5"
    ):
        class TestNWBConverter(NWBConverter):
            data_interface_classes = dict(Test=type(self.interface))

        test_kwargs = self.test_kwargs[0] if isinstance(self.test_kwargs, list) else self.test_kwargs
        source_data = dict(Test=test_kwargs)
        converter = TestNWBConverter(source_data=source_data)

        conversion_options = dict(Test=self.conversion_options)
        converter.run_conversion(
            nwbfile_path=nwbfile_path,
            overwrite=True,
            metadata=metadata,
            backend=backend,
            conversion_options=conversion_options,
        )

    def check_run_conversion_in_nwbconverter_with_backend_configuration(
        self, nwbfile_path: str, metadata: dict, backend: Union["hdf5", "zarr"] = "hdf5"
    ):
        class TestNWBConverter(NWBConverter):
            data_interface_classes = dict(Test=type(self.interface))

        test_kwargs = self.test_kwargs[0] if isinstance(self.test_kwargs, list) else self.test_kwargs
        source_data = dict(Test=test_kwargs)
        converter = TestNWBConverter(source_data=source_data)

        conversion_options = dict(Test=self.conversion_options)

        nwbfile = converter.create_nwbfile(metadata=metadata, conversion_options=conversion_options)
        backend_configuration = converter.get_default_backend_configuration(nwbfile=nwbfile, backend=backend)
        converter.run_conversion(
            nwbfile_path=nwbfile_path,
            nwbfile=nwbfile,
            overwrite=True,
            metadata=metadata,
            backend_configuration=backend_configuration,
            conversion_options=conversion_options,
        )

    def test_all_conversion_checks(self, metadata: dict):
        interface_kwargs = self.interface_kwargs
        if isinstance(interface_kwargs, dict):
            interface_kwargs = [interface_kwargs]
        for num, kwargs in enumerate(interface_kwargs):
            with self.subTest(str(num)):
                self.case = num
                self.test_kwargs = kwargs
                self.interface = self.data_interface_cls(**self.test_kwargs)

                self.check_metadata_schema_valid()
                self.check_conversion_options_schema_valid()
                self.check_metadata()
                self.nwbfile_path = str(self.save_directory / f"{self.__class__.__name__}_{num}.nwb")

                self.check_no_metadata_mutation(metadata=metadata)

                self.check_configure_backend_for_equivalent_nwbfiles(metadata=metadata)

                self.check_run_conversion_in_nwbconverter_with_backend(
                    nwbfile_path=self.nwbfile_path, metadata=metadata, backend="hdf5"
                )
                self.check_run_conversion_in_nwbconverter_with_backend_configuration(
                    nwbfile_path=self.nwbfile_path, metadata=metadata, backend="hdf5"
                )

                self.check_run_conversion_with_backend(
                    nwbfile_path=self.nwbfile_path, metadata=metadata, backend="hdf5"
                )
                self.check_run_conversion_with_backend_configuration(
                    nwbfile_path=self.nwbfile_path, metadata=metadata, backend="hdf5"
                )

                self.check_read_nwb(nwbfile_path=self.nwbfile_path)

                # TODO: enable when all H5DataIO prewraps are gone
                # self.nwbfile_path = str(self.save_directory / f"{self.__class__.__name__}_{num}.nwb.zarr")
                # self.check_run_conversion(nwbfile_path=self.nwbfile_path, backend="zarr")
                # self.check_run_conversion_custom_backend(nwbfile_path=self.nwbfile_path, backend="zarr")
                # self.check_basic_zarr_read(nwbfile_path=self.nwbfile_path)

                # Any extra custom checks to run
                self.run_custom_checks()

    def check_interface_get_original_timestamps(self):
        """
        Just to ensure each interface can call .get_original_timestamps() without an error raising.

        Also, that it always returns non-empty.
        """
        self.setUpFreshInterface()
        t1 = self.conversion_options.get("t1", 0.0)
        t2 = self.conversion_options.get("t2", 0.0)
        stream_name_to_timestamps = self.interface.get_original_timestamps(t1=t1, t2=t2)
        for stream_name, timestamps in stream_name_to_timestamps.items():
            assert len(timestamps) != 0, f"Timestamps for {stream_name} are empty."

    def check_interface_get_timestamps(self):
        """
        Just to ensure each interface can call .get_timestamps() without an error raising.

        Also, that it always returns non-empty.
        """
        self.setUpFreshInterface()
        t1 = self.conversion_options.get("t1", 0.0)
        t2 = self.conversion_options.get("t2", 0.0)
        stream_name_to_timestamps = self.interface.get_timestamps(t1=t1, t2=t2)
        for stream_name, timestamps in stream_name_to_timestamps.items():
            assert len(timestamps) != 0, f"Timestamps for {stream_name} are empty."

    def check_interface_set_aligned_timestamps(self):
        """Ensure that internal mechanisms for the timestamps getter/setter work as expected."""
        t1 = self.conversion_options.get("t1", 0.0)
        t2 = self.conversion_options.get("t2", 0.0)
        self.setUpFreshInterface()
        unaligned_stream_name_to_timestamps = self.interface.get_original_timestamps(t1=t1, t2=t2)

        random_number_generator = np.random.default_rng(seed=0)
        aligned_stream_name_to_timestamps = {}
        for stream_name, unaligned_timestamps in unaligned_stream_name_to_timestamps.items():
            aligned_timestamps = (
                unaligned_timestamps + 1.23 + random_number_generator.random(size=unaligned_timestamps.shape)
            )
            aligned_stream_name_to_timestamps[stream_name] = aligned_timestamps
        self.interface.set_aligned_timestamps(stream_name_to_aligned_timestamps=aligned_stream_name_to_timestamps)
        t1 += 1.23 if t1 != 0.0 else 0.0
        t2 += 2.23 if t2 != 0.0 else 0.0

        retrieved_aligned_stream_name_to_timestamps = self.interface.get_timestamps(t1=t1, t2=t2)
        for stream_name, aligned_timestamps in aligned_stream_name_to_timestamps.items():
            retrieved_aligned_timestamps = retrieved_aligned_stream_name_to_timestamps[stream_name]
            assert_array_equal(retrieved_aligned_timestamps, aligned_timestamps)

    def check_shift_timestamps_by_start_time(self):
        """Ensure that internal mechanisms for shifting timestamps by a starting time work as expected."""
        t1 = self.conversion_options.get("t1", 0.0)
        t2 = self.conversion_options.get("t2", 0.0)
        self.setUpFreshInterface()
        unaligned_stream_name_to_timestamps = self.interface.get_original_timestamps(t1=t1, t2=t2)

        aligned_starting_time = 1.23
        self.interface.set_aligned_starting_time(aligned_starting_time=aligned_starting_time, t1=t1, t2=t2)
        t1 += aligned_starting_time if t1 != 0.0 else 0.0
        t2 += aligned_starting_time if t2 != 0.0 else 0.0

        aligned_stream_name_to_timestamps = self.interface.get_timestamps(t1=t1, t2=t2)
        expected_timestamps_dict = {
            name: unaligned_timestamps + aligned_starting_time
            for name, unaligned_timestamps in unaligned_stream_name_to_timestamps.items()
        }
        for name, expected_timestamps in expected_timestamps_dict.items():
            timestamps = aligned_stream_name_to_timestamps[name]
            assert_array_equal(timestamps, expected_timestamps)

    def check_interface_original_timestamps_inmutability(self):
        """Check aligning the timestamps for the interface does not change the value of .get_original_timestamps()."""
        t1 = self.conversion_options.get("t1", 0.0)
        t2 = self.conversion_options.get("t2", 0.0)
        self.setUpFreshInterface()
        pre_alignment_stream_name_to_timestamps = self.interface.get_original_timestamps(t1=t1, t2=t2)

        aligned_stream_name_to_timestamps = {
            name: pre_alignment_timestamps + 1.23
            for name, pre_alignment_timestamps in pre_alignment_stream_name_to_timestamps.items()
        }
        self.interface.set_aligned_timestamps(stream_name_to_aligned_timestamps=aligned_stream_name_to_timestamps)

        post_alignment_stream_name_to_timestamps = self.interface.get_original_timestamps(t1=t1, t2=t2)
        for name, post_alignment_timestamps in post_alignment_stream_name_to_timestamps.items():
            pre_alignment_timestamps = pre_alignment_stream_name_to_timestamps[name]
            assert_array_equal(post_alignment_timestamps, pre_alignment_timestamps)

    def test_interface_alignment(self):
        interface_kwargs = self.interface_kwargs
        if isinstance(interface_kwargs, dict):
            interface_kwargs = [interface_kwargs]
        for num, kwargs in enumerate(interface_kwargs):
            with self.subTest(str(num)):
                self.case = num
                self.test_kwargs = kwargs

                self.check_interface_get_original_timestamps()
                self.check_interface_get_timestamps()
                self.check_interface_set_aligned_timestamps()
                self.check_shift_timestamps_by_start_time()
                self.check_interface_original_timestamps_inmutability()
                self.check_nwbfile_temporal_alignment()



"""Collection of helper functions related to configuration of datasets dependent on backend."""

from typing import Generator, Literal, Union

import h5py
import numpy as np
import zarr
from hdmf import Container
from hdmf.data_utils import DataIO
from hdmf.utils import get_data_shape
from hdmf_zarr import NWBZarrIO
from pynwb import NWBHDF5IO, NWBFile
from pynwb.base import DynamicTable, TimeSeriesReferenceVectorData
from pynwb.file import NWBContainer

from ._configuration_models import DATASET_IO_CONFIGURATIONS
from ._configuration_models._base_dataset_io import DatasetIOConfiguration


def _get_io_mode(io: Union[NWBHDF5IO, NWBZarrIO]) -> str:
    """NWBHDF5IO and NWBZarrIO have different ways of storing the io mode (e.g. "r", "a", "w") they used on a path."""
    if isinstance(io, NWBHDF5IO):
        return io.mode
    elif isinstance(io, NWBZarrIO):
        return io._ZarrIO__mode


def _is_dataset_written_to_file(
    candidate_dataset: Union[h5py.Dataset, zarr.Array],
    backend: Literal["hdf5", "zarr"],
    existing_file: Union[h5py.File, zarr.Group, None],
) -> bool:
    """
    Determine if the neurodata object is already written to the file on disk.

    This object should then be skipped by the `get_io_datasets` function when working in append mode.
    """
    if existing_file is None:
        return False

    return (
        isinstance(candidate_dataset, h5py.Dataset)  # If the source data is an HDF5 Dataset
        and backend == "hdf5"
        and candidate_dataset.file == existing_file  # If the source HDF5 Dataset is the appending NWBFile
    ) or (
        isinstance(candidate_dataset, zarr.Array)  # If the source data is a Zarr Array
        and backend == "zarr"
        and candidate_dataset.store == existing_file  # If the source Zarr 'file' is the appending NWBFile
    )


def get_default_dataset_io_configurations(
    nwbfile: NWBFile,
    backend: Union[None, Literal["hdf5", "zarr"]] = None,  # None for auto-detect from append mode, otherwise required
) -> Generator[DatasetIOConfiguration, None, None]:
    """
    Generate DatasetIOConfiguration objects for wrapping NWB file objects with a specific backend.

    This method automatically detects all objects in an NWB file that can be wrapped in a hdmf.DataIO.
    If the NWB file is in append mode, it supports auto-detection of the backend.
    Otherwise, it requires a backend specification.

    Parameters
    ----------
    nwbfile : pynwb.NWBFile
        An in-memory NWBFile object, either generated from the base class or read from an existing file of any backend.
    backend : "hdf5" or "zarr"
        Which backend format type you would like to use in configuring each dataset's compression methods and options.

    Yields
    ------
    DatasetIOConfiguration
        A summary of each detected object that can be wrapped in a hdmf.DataIO.
    """

    DatasetIOConfigurationClass = DATASET_IO_CONFIGURATIONS[backend]

    if backend is None and nwbfile.read_io is None:
        raise ValueError(
            "Keyword argument `backend` (either 'hdf5' or 'zarr') must be specified if the `nwbfile` was not "
            "read from an existing file!"
        )
    if backend is None and nwbfile.read_io is not None and nwbfile.read_io.mode not in ("r+", "a"):
        raise ValueError(
            "Keyword argument `backend` (either 'hdf5' or 'zarr') must be specified if the `nwbfile` is being appended."
        )

    detected_backend = None
    existing_file = None
    if isinstance(nwbfile.read_io, NWBHDF5IO) and _get_io_mode(io=nwbfile.read_io) in ("r+", "a"):
        detected_backend = "hdf5"
        existing_file = nwbfile.read_io._file
    elif isinstance(nwbfile.read_io, NWBZarrIO) and _get_io_mode(io=nwbfile.read_io) in ("r+", "a"):
        detected_backend = "zarr"
        existing_file = nwbfile.read_io.file.store
    backend = backend or detected_backend

    if detected_backend is not None and detected_backend != backend:
        raise ValueError(
            f"Detected backend '{detected_backend}' for appending file, but specified `backend` "
            f"({backend}) does not match! Set `backend=None` or remove the keyword argument to allow it to auto-detect."
        )

    known_dataset_fields = ("data", "timestamps")
    for neurodata_object in nwbfile.objects.values():
        if isinstance(neurodata_object, DynamicTable):
            dynamic_table = neurodata_object  # For readability

            for column in dynamic_table.columns:
                candidate_dataset = column.data  # VectorData object
                # noinspection PyTypeChecker
                if _is_dataset_written_to_file(
                    candidate_dataset=candidate_dataset, backend=backend, existing_file=existing_file
                ):
                    continue  # Skip

                # Skip over columns that are already wrapped in DataIO
                if isinstance(candidate_dataset, DataIO):
                    continue  # Skip

                # Skip over columns whose values are links, such as the 'group' of an ElectrodesTable
                if any(isinstance(value, Container) for value in candidate_dataset):
                    continue  # Skip

                # Skip when columns whose values are a reference type
                if isinstance(column, TimeSeriesReferenceVectorData):
                    continue

                # Skip datasets with any zero-length axes
                dataset_name = "data"
                candidate_dataset = getattr(column, dataset_name)
                full_shape = get_data_shape(data=candidate_dataset)
                if any(axis_length == 0 for axis_length in full_shape):
                    continue

                dataset_io_configuration = DatasetIOConfigurationClass.from_neurodata_object(
                    neurodata_object=column, dataset_name=dataset_name
                )

                yield dataset_io_configuration
        elif isinstance(neurodata_object, NWBContainer):
            for known_dataset_field in known_dataset_fields:
                # Skip optional fields that aren't present
                if known_dataset_field not in neurodata_object.fields:
                    continue

                candidate_dataset = getattr(neurodata_object, known_dataset_field)

                # Skip if already written to file
                # noinspection PyTypeChecker
                if _is_dataset_written_to_file(
                    candidate_dataset=candidate_dataset, backend=backend, existing_file=existing_file
                ):
                    continue

                # Skip over datasets that are already wrapped in DataIO
                if isinstance(candidate_dataset, DataIO):
                    continue

                # Skip edge case of in-memory ImageSeries with external mode; data is in fields and is empty array
                if isinstance(candidate_dataset, np.ndarray) and candidate_dataset.size == 0:
                    continue

                # Skip datasets with any zero-length axes
                candidate_dataset = getattr(neurodata_object, known_dataset_field)
                full_shape = get_data_shape(data=candidate_dataset)
                if any(axis_length == 0 for axis_length in full_shape):
                    continue

                dataset_io_configuration = DatasetIOConfigurationClass.from_neurodata_object(
                    neurodata_object=neurodata_object, dataset_name=known_dataset_field
                )

                yield dataset_io_configuration



import re
from pathlib import Path, PureWindowsPath
from typing import Union


def infer_path(path: str) -> Union[PureWindowsPath, Path]:
    """
    Infers and returns the appropriate path object based on the path string.

    Parameters
    ----------
    path : str
        The file path string to be inferred.

    Returns
    -------
    Union[PureWindowsPath, Path]
        Returns a ``PureWindowsPath`` object if the string is in Windows format,
        else returns a ``Path`` object.

    Examples
    --------
    >>> infer_path(r'C:\\Users\\anon\\Desktop\\file.txt')
    PureWindowsPath('C:/Users/anon/Desktop/file.txt')

    >>> infer_path('/Users/anon/Desktop/file.txt')
    PosixPath('/Users/anon/Desktop/file.txt')
    """
    if re.match(r"^[A-Za-z]:\\", path) or "\\" in path:
        return PureWindowsPath(path)
    else:
        return Path(path)



from pydantic import DirectoryPath, validate_call

from ..baserecordingextractorinterface import BaseRecordingExtractorInterface


class TdtRecordingInterface(BaseRecordingExtractorInterface):
    """Primary data interface class for converting Tucker-Davis Technologies (TDT) data."""

    display_name = "TDT Recording"
    associated_suffixes = (".tbk", ".tbx", ".tev", ".tsq")
    info = "Interface for TDT recording data."

    @validate_call
    def __init__(
        self,
        folder_path: DirectoryPath,
        gain: float,
        stream_id: str = "0",
        verbose: bool = True,
        es_key: str = "ElectricalSeries",
    ):
        """
        Initialize reading of a TDT recording.

        Parameters
        ----------
        folder_path : str or Path
            Path to the directory with the corresponding files (TSQ, TBK, TEV, SEV)
        stream_id : str, "0" by default
            Select from multiple streams.
        gain : float
            The conversion factor from int16 to microvolts.
        verbose : bool, default: True
            Allows verbose.
        es_key : str, optional


        Notes
        -----
        Stream "0" corresponds to LFP for gin data. Other streams seem non-electrical.
        """
        super().__init__(
            folder_path=folder_path,
            stream_id=stream_id,
            verbose=verbose,
            es_key=es_key,
        )

        # Fix channel name format
        channel_names = self.recording_extractor.get_property("channel_name")
        channel_names = [name.replace("'", "")[1:] for name in channel_names]
        self.recording_extractor.set_property(key="channel_name", values=channel_names)

        self.recording_extractor.set_channel_gains(gain)



from typing import Any, Iterable, Literal, Union

import h5py
import numcodecs
import numpy as np

from ...nwb_helpers import (
    AVAILABLE_HDF5_COMPRESSION_METHODS,
    AVAILABLE_ZARR_COMPRESSION_METHODS,
    HDF5BackendConfiguration,
    HDF5DatasetIOConfiguration,
    ZarrBackendConfiguration,
    ZarrDatasetIOConfiguration,
)


def mock_HDF5DatasetIOConfiguration(
    object_id: str = "481a0860-3a0c-40ec-b931-df4a3e9b101f",
    location_in_file: str = "acquisition/TestElectricalSeries/data",
    dataset_name: Literal["data", "timestamps"] = "data",
    full_shape: tuple[int, ...] = (60 * 30_000, 384),  # ~1 minute of v1 NeuroPixels probe
    dtype: np.dtype = np.dtype("int16"),
    chunk_shape: tuple[int, ...] = (78_125, 64),  # ~10 MB
    buffer_shape: tuple[int, ...] = (1_250_000, 384),  # ~1 GB
    compression_method: Union[
        Literal[tuple(AVAILABLE_HDF5_COMPRESSION_METHODS.keys())], h5py._hl.filters.FilterRefBase, None
    ] = "gzip",
    compression_options: Union[dict[str, Any], None] = None,
) -> HDF5DatasetIOConfiguration:
    """Mock object of a HDF5DatasetIOConfiguration with NeuroPixel-like values to show chunk/buffer recommendations."""
    return HDF5DatasetIOConfiguration(
        object_id=object_id,
        location_in_file=location_in_file,
        dataset_name=dataset_name,
        full_shape=full_shape,
        dtype=dtype,
        chunk_shape=chunk_shape,
        buffer_shape=buffer_shape,
        compression_method=compression_method,
        compression_options=compression_options,
    )


def mock_ZarrDatasetIOConfiguration(
    object_id: str = "481a0860-3a0c-40ec-b931-df4a3e9b101f",
    location_in_file: str = "acquisition/TestElectricalSeries/data",
    dataset_name: Literal["data", "timestamps"] = "data",
    full_shape: tuple[int, ...] = (60 * 30_000, 384),  # ~1 minute of v1 NeuroPixels probe
    dtype: np.dtype = np.dtype("int16"),
    chunk_shape: tuple[int, ...] = (78_125, 64),  # ~10 MB
    buffer_shape: tuple[int, ...] = (1_250_000, 384),  # ~1 GB
    compression_method: Union[
        Literal[tuple(AVAILABLE_ZARR_COMPRESSION_METHODS.keys())], numcodecs.abc.Codec, None
    ] = "gzip",
    compression_options: Union[dict[str, Any]] = None,
    filter_methods: Iterable[
        Union[Literal[tuple(AVAILABLE_ZARR_COMPRESSION_METHODS.keys())], numcodecs.abc.Codec, None]
    ] = None,
    filter_options: Union[Iterable[dict[str, Any]], None] = None,
) -> ZarrDatasetIOConfiguration:
    """Mock object of a ZarrDatasetIOConfiguration with NeuroPixel-like values to show chunk/buffer recommendations."""
    return ZarrDatasetIOConfiguration(
        object_id=object_id,
        location_in_file=location_in_file,
        dataset_name=dataset_name,
        full_shape=full_shape,
        dtype=dtype,
        chunk_shape=chunk_shape,
        buffer_shape=buffer_shape,
        compression_method=compression_method,
        compression_options=compression_options,
        filter_methods=filter_methods,
        filter_options=filter_options,
    )


def mock_HDF5BackendConfiguration() -> HDF5BackendConfiguration:
    """Mock instance of a HDF5BackendConfiguration with two NeuroPixel-like datasets."""
    dataset_configurations: dict[str, HDF5DatasetIOConfiguration] = {
        "acquisition/TestElectricalSeriesAP/data": mock_HDF5DatasetIOConfiguration(
            location_in_file="acquisition/TestElectricalSeriesAP/data", dataset_name="data"
        ),
        "acquisition/TestElectricalSeriesLF/data": mock_HDF5DatasetIOConfiguration(
            object_id="bc37e164-519f-4b65-a976-206440f1d325",
            location_in_file="acquisition/TestElectricalSeriesLF/data",
            dataset_name="data",
            full_shape=(75_000, 384),
            chunk_shape=(37_500, 128),  # ~10 MB
            buffer_shape=(75_000, 384),
        ),
    }

    return HDF5BackendConfiguration(
        dataset_configurations=dataset_configurations,
    )


def mock_ZarrBackendConfiguration() -> ZarrBackendConfiguration:
    """Mock instance of a HDF5BackendConfiguration with several NeuroPixel-like datasets."""
    dataset_configurations: dict[str, ZarrDatasetIOConfiguration] = {
        "acquisition/TestElectricalSeriesAP/data": mock_ZarrDatasetIOConfiguration(
            location_in_file="acquisition/TestElectricalSeriesAP/data",
            dataset_name="data",
            filter_methods=["delta"],
        ),
        "acquisition/TestElectricalSeriesLF/data": mock_ZarrDatasetIOConfiguration(
            object_id="bc37e164-519f-4b65-a976-206440f1d325",
            location_in_file="acquisition/TestElectricalSeriesLF/data",
            dataset_name="data",
            full_shape=(75_000, 384),
            chunk_shape=(37_500, 128),  # ~10 MB
            buffer_shape=(75_000, 384),
            filter_methods=["delta"],
        ),
    }

    return ZarrBackendConfiguration(dataset_configurations=dataset_configurations)



import importlib
import json
import uuid
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Literal, Optional, Union

from jsonschema.validators import validate
from pydantic import FilePath, validate_call
from pynwb import NWBFile

from .tools.nwb_helpers import (
    HDF5BackendConfiguration,
    ZarrBackendConfiguration,
    configure_backend,
    get_default_backend_configuration,
    make_nwbfile_from_metadata,
    make_or_load_nwbfile,
)
from .tools.nwb_helpers._metadata_and_file_helpers import _resolve_backend
from .utils import (
    _NWBMetaDataEncoder,
    get_json_schema_from_method_signature,
    load_dict_from_file,
)
from .utils.dict import DeepDict


class BaseDataInterface(ABC):
    """Abstract class defining the structure of all DataInterfaces."""

    display_name: Union[str, None] = None
    keywords: tuple[str] = tuple()
    associated_suffixes: tuple[str] = tuple()
    info: Union[str, None] = None

    @classmethod
    def get_source_schema(cls) -> dict:
        """Infer the JSON schema for the source_data from the method signature (annotation typing)."""
        return get_json_schema_from_method_signature(cls, exclude=["source_data"])

    @validate_call
    def __init__(self, verbose: bool = False, **source_data):
        self.verbose = verbose
        self.source_data = source_data

    def get_metadata_schema(self) -> dict:
        """Retrieve JSON schema for metadata."""
        metadata_schema = load_dict_from_file(Path(__file__).parent / "schemas" / "base_metadata_schema.json")
        return metadata_schema

    def get_metadata(self) -> DeepDict:
        """Child DataInterface classes should override this to match their metadata."""
        metadata = DeepDict()
        metadata["NWBFile"]["session_description"] = ""
        metadata["NWBFile"]["identifier"] = str(uuid.uuid4())

        # Add NeuroConv watermark (overridden if going through the GUIDE)
        neuroconv_version = importlib.metadata.version("neuroconv")
        metadata["NWBFile"]["source_script"] = f"Created using NeuroConv v{neuroconv_version}"
        metadata["NWBFile"]["source_script_file_name"] = __file__  # Required for validation

        return metadata

    def validate_metadata(self, metadata: dict, append_mode: bool = False) -> None:
        """Validate the metadata against the schema."""
        encoder = _NWBMetaDataEncoder()
        # The encoder produces a serialized object, so we deserialized it for comparison

        serialized_metadata = encoder.encode(metadata)
        decoded_metadata = json.loads(serialized_metadata)
        metdata_schema = self.get_metadata_schema()
        if append_mode:
            # Eliminate required from NWBFile
            nwbfile_schema = metdata_schema["properties"]["NWBFile"]
            nwbfile_schema.pop("required", None)

        validate(instance=decoded_metadata, schema=metdata_schema)

    def get_conversion_options_schema(self) -> dict:
        """Infer the JSON schema for the conversion options from the method signature (annotation typing)."""
        return get_json_schema_from_method_signature(self.add_to_nwbfile, exclude=["nwbfile", "metadata"])

    def create_nwbfile(self, metadata: Optional[dict] = None, **conversion_options) -> NWBFile:
        """
        Create and return an in-memory pynwb.NWBFile object with this interface's data added to it.

        Parameters
        ----------
        metadata : dict, optional
            Metadata dictionary with information used to create the NWBFile.
        **conversion_options
            Additional keyword arguments to pass to the `.add_to_nwbfile` method.

        Returns
        -------
        nwbfile : pynwb.NWBFile
            The in-memory object with this interface's data added to it.
        """
        if metadata is None:
            metadata = self.get_metadata()

        nwbfile = make_nwbfile_from_metadata(metadata=metadata)
        self.add_to_nwbfile(nwbfile=nwbfile, metadata=metadata, **conversion_options)

        return nwbfile

    @abstractmethod
    def add_to_nwbfile(self, nwbfile: NWBFile, **conversion_options) -> None:
        """
        Define a protocol for mapping the data from this interface to NWB neurodata objects.

        These neurodata objects should also be added to the in-memory pynwb.NWBFile object in this step.

        Parameters
        ----------
        nwbfile : pynwb.NWBFile
            The in-memory object to add the data to.
        **conversion_options
            Additional keyword arguments to pass to the `.add_to_nwbfile` method.
        """
        raise NotImplementedError

    def run_conversion(
        self,
        nwbfile_path: FilePath,
        nwbfile: Optional[NWBFile] = None,
        metadata: Optional[dict] = None,
        overwrite: bool = False,
        backend: Optional[Literal["hdf5", "zarr"]] = None,
        backend_configuration: Optional[Union[HDF5BackendConfiguration, ZarrBackendConfiguration]] = None,
        **conversion_options,
    ):
        """
        Run the NWB conversion for the instantiated data interface.

        Parameters
        ----------
        nwbfile_path : FilePathType
            Path for where the data will be written or appended.
        nwbfile : NWBFile, optional
            An in-memory NWBFile object to write to the location.
        metadata : dict, optional
            Metadata dictionary with information used to create the NWBFile when one does not exist or overwrite=True.
        overwrite : bool, default: False
            Whether to overwrite the NWBFile if one exists at the nwbfile_path.
            The default is False (append mode).
        backend : {"hdf5", "zarr"}, optional
            The type of backend to use when writing the file.
            If a `backend_configuration` is not specified, the default type will be "hdf5".
            If a `backend_configuration` is specified, then the type will be auto-detected.
        backend_configuration : HDF5BackendConfiguration or ZarrBackendConfiguration, optional
            The configuration model to use when configuring the datasets for this backend.
            To customize, call the `.get_default_backend_configuration(...)` method, modify the returned
            BackendConfiguration object, and pass that instead.
            Otherwise, all datasets will use default configuration settings.
        """

        backend = _resolve_backend(backend, backend_configuration)
        no_nwbfile_provided = nwbfile is None  # Otherwise, variable reference may mutate later on inside the context

        if metadata is None:
            metadata = self.get_metadata()

        file_initially_exists = Path(nwbfile_path).exists() if nwbfile_path is not None else False
        append_mode = file_initially_exists and not overwrite

        self.validate_metadata(metadata=metadata, append_mode=append_mode)

        with make_or_load_nwbfile(
            nwbfile_path=nwbfile_path,
            nwbfile=nwbfile,
            metadata=metadata,
            overwrite=overwrite,
            backend=backend,
            verbose=getattr(self, "verbose", False),
        ) as nwbfile_out:
            if no_nwbfile_provided:
                self.add_to_nwbfile(nwbfile=nwbfile_out, metadata=metadata, **conversion_options)

            if backend_configuration is None:
                backend_configuration = self.get_default_backend_configuration(nwbfile=nwbfile_out, backend=backend)

            configure_backend(nwbfile=nwbfile_out, backend_configuration=backend_configuration)

    @staticmethod
    def get_default_backend_configuration(
        nwbfile: NWBFile,
        # TODO: when all H5DataIO prewraps are gone, introduce Zarr safely
        # backend: Union[Literal["hdf5", "zarr"]],
        backend: Literal["hdf5"] = "hdf5",
    ) -> Union[HDF5BackendConfiguration, ZarrBackendConfiguration]:
        """
        Fill and return a default backend configuration to serve as a starting point for further customization.

        Parameters
        ----------
        nwbfile : pynwb.NWBFile
            The in-memory object with this interface's data already added to it.
        backend : "hdf5", default: "hdf5"
            The type of backend to use when creating the file.
            Additional backend types will be added soon.

        Returns
        -------
        backend_configuration : HDF5BackendConfiguration or ZarrBackendConfiguration
            The default configuration for the specified backend type.
        """
        return get_default_backend_configuration(nwbfile=nwbfile, backend=backend)



from abc import abstractmethod
from pathlib import Path
from typing import Optional

import numpy as np
from pydantic import FilePath, validate_call
from pynwb import NWBFile

from ...basedatainterface import BaseDataInterface
from ...tools.text import convert_df_to_time_intervals
from ...utils.dict import load_dict_from_file


class TimeIntervalsInterface(BaseDataInterface):
    """Abstract Interface for time intervals."""

    keywords = ("table", "trials", "epochs", "time intervals")

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        read_kwargs: Optional[dict] = None,
        verbose: bool = True,
    ):
        """
        Parameters
        ----------
        file_path : FilePath
        read_kwargs : dict, optional
        verbose : bool, default: True
        """
        read_kwargs = read_kwargs or dict()
        super().__init__(file_path=file_path)
        self.verbose = verbose

        self._read_kwargs = read_kwargs
        self.dataframe = self._read_file(file_path, **read_kwargs)
        self.time_intervals = None

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()
        metadata["TimeIntervals"] = dict(
            trials=dict(
                table_name="trials",
                table_description=f"experimental trials generated from {self.source_data['file_path']}",
            )
        )

        return metadata

    def get_metadata_schema(self) -> dict:
        fpath = Path(__file__).parent.parent.parent / "schemas" / "timeintervals_schema.json"
        return load_dict_from_file(fpath)

    def get_original_timestamps(self, column: str) -> np.ndarray:
        if not column.endswith("_time"):
            raise ValueError("Timing columns on a TimeIntervals table need to end with '_time'!")

        return self._read_file(**self.source_data, **self._read_kwargs)[column].values

    def get_timestamps(self, column: str) -> np.ndarray:
        if not column.endswith("_time"):
            raise ValueError("Timing columns on a TimeIntervals table need to end with '_time'!")

        return self.dataframe[column].values

    def set_aligned_starting_time(self, aligned_starting_time: float):
        timing_columns = [column for column in self.dataframe.columns if column.endswith("_time")]

        for column in timing_columns:
            self.dataframe[column] += aligned_starting_time

    def set_aligned_timestamps(
        self, aligned_timestamps: np.ndarray, column: str, interpolate_other_columns: bool = False
    ):
        if not column.endswith("_time"):
            raise ValueError("Timing columns on a TimeIntervals table need to end with '_time'!")

        unaligned_timestamps = np.array(self.dataframe[column])
        self.dataframe[column] = aligned_timestamps

        if not interpolate_other_columns:
            return

        other_timing_columns = [
            other_column
            for other_column in self.dataframe.columns
            if other_column.endswith("_time") and other_column != column
        ]
        for other_timing_column in other_timing_columns:
            self.align_by_interpolation(
                unaligned_timestamps=unaligned_timestamps,
                aligned_timestamps=aligned_timestamps,
                column=other_timing_column,
            )

    def align_by_interpolation(self, unaligned_timestamps: np.ndarray, aligned_timestamps: np.ndarray, column: str):
        current_timestamps = self.get_timestamps(column=column)
        assert (
            current_timestamps[1] >= unaligned_timestamps[0]
        ), "All current timestamps except for the first must be strictly within the unaligned mapping."
        assert (
            current_timestamps[-2] <= unaligned_timestamps[-1]
        ), "All current timestamps except for the last must be strictly within the unaligned mapping."
        # Assume timing column is ascending otherwise

        self.set_aligned_timestamps(
            aligned_timestamps=np.interp(
                x=current_timestamps,
                xp=unaligned_timestamps,
                fp=aligned_timestamps,
                left=2 * aligned_timestamps[0] - aligned_timestamps[1],  # If first or last values are outside alignment
                right=2 * aligned_timestamps[-1] - aligned_timestamps[-2],  # then use the most recent diff to regress
            ),
            column=column,
        )

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: Optional[dict] = None,
        tag: str = "trials",
        column_name_mapping: dict[str, str] = None,
        column_descriptions: dict[str, str] = None,
    ) -> NWBFile:
        """
        Run the NWB conversion for the instantiated data interface.

        Parameters
        ----------
        nwbfile : NWBFile, optional
            An in-memory NWBFile object to write to the location.
        metadata : dict, optional
            Metadata dictionary with information used to create the NWBFile when one does not exist or overwrite=True.
        tag : str, default: "trials"
        column_name_mapping: dict, optional
            If passed, rename subset of columns from key to value.
        column_descriptions: dict, optional
            Keys are the names of the columns (after renaming) and values are the descriptions. If not passed,
            the names of the columns are used as descriptions.

        """
        metadata = metadata or self.get_metadata()
        self.time_intervals = convert_df_to_time_intervals(
            self.dataframe,
            column_name_mapping=column_name_mapping,
            column_descriptions=column_descriptions,
            **metadata["TimeIntervals"][tag],
        )
        nwbfile.add_time_intervals(self.time_intervals)

        return nwbfile

    @abstractmethod
    def _read_file(self, file_path: FilePath, **read_kwargs):
        pass



import numpy as np
from pydantic import FilePath


def get_medpc_variables(file_path: FilePath, variable_names: list) -> dict:
    """
    Get the values of the given single-line variables from a MedPC file for all sessions in that file.

    Parameters
    ----------
    file_path : FilePathType
        The path to the MedPC file.
    variable_names : list
        The names of the variables to get the values of.

    Returns
    -------
    dict
        A dictionary with the variable names as keys and a list of variable values as values.
    """
    with open(file_path, "r") as f:
        lines = f.readlines()
    medpc_variables = {name: [] for name in variable_names}
    for line in lines:
        for variable_name in variable_names:
            if line.startswith(variable_name):
                medpc_variables[variable_name].append(line.split(":", maxsplit=1)[1].strip())
    return medpc_variables


def _get_session_lines(lines: list, session_conditions: dict, start_variable: str) -> list:
    """
    Get the lines for a session from a MedPC file.

    Parameters
    ----------
    lines : list
        The lines of the MedPC file.
    session_conditions : dict
        The conditions that define the session. The keys are the names of the single-line variables (ex. 'Start Date')
        and the values are the values of those variables for the desired session (ex. '11/09/18').
    start_variable : str
        The name of the variable that starts the session (ex. 'Start Date').

    Returns
    -------
    list
        The lines for the session.

    Raises
    ------
    ValueError
        If the session with the given conditions could not be found.
    ValueError
        If the start variable of the session with the given conditions could not be found.

    Notes
    -----
    If multiple sessions satisfy the session_conditions, the first session that meets the conditions will be returned.
    """
    session_condition_has_been_met = {name: False for name in session_conditions}
    start_line, end_line = None, len(lines)
    for i, line in enumerate(lines):
        line = line.strip()
        if line.startswith(f"{start_variable}:"):
            start_line = i
        for condition_name, condition_value in session_conditions.items():
            if line == f"{condition_name}: {condition_value}":
                session_condition_has_been_met[condition_name] = True
        if line == "" and all(session_condition_has_been_met.values()):
            end_line = i
            break
        elif line == "":
            session_condition_has_been_met = {name: False for name in session_conditions}
            start_line = None
    if not all(session_condition_has_been_met.values()):
        raise ValueError(f"Could not find the session with conditions {session_conditions}")
    if start_line is None:
        raise ValueError(
            f"Could not find the start variable ({start_variable}) of the session with conditions {session_conditions}"
        )
    session_lines = lines[start_line:end_line]
    return session_lines


def read_medpc_file(
    file_path: FilePath,
    medpc_name_to_info_dict: dict,
    session_conditions: dict,
    start_variable: str,
) -> dict:
    """
    Read a raw MedPC text file into a dictionary.

    Parameters
    ----------
    file_path : FilePathType
        The path to the MedPC file.
    medpc_name_to_info_dict : dict
        A dictionary where the keys are the MedPC variable names and the values are dictionaries with the keys 'name' and
        'is_array'. 'name' is the name of the variable in the output dictionary and 'is_array' is a boolean indicating
        whether the variable is an array.  Ex. {'Start Date': {'name': 'start_date', 'is_array': False}}
    session_conditions : dict
        The conditions that define the session. The keys are the names of the single-line variables (ex. 'Start Date')
        and the values are the values of those variables for the desired session (ex. '11/09/18').
    start_variable : str
        The name of the variable that starts the session (ex. 'Start Date').

    Returns
    -------
    dict
        A dictionary with the variable names as keys and the data extracted from medpc output are the values.

    Raises
    ------
    ValueError
        If the session with the given conditions could not be found.
    """
    with open(file_path, "r") as f:
        lines = f.readlines()
    session_lines = _get_session_lines(lines, session_conditions=session_conditions, start_variable=start_variable)

    # Parse the session lines into a dictionary
    session_dict = {}
    for i, line in enumerate(session_lines):
        line = line.rstrip()
        if line.startswith("\\"):  # \\ indicates a commented line in the MedPC file
            continue
        assert ":" in line, f"Could not find ':' in line {repr(line)}"
        split_line = line.split(":", maxsplit=1)
        medpc_name, data = split_line
        data = data.strip()
        if "\t" in data:  # some sessions have a bunch of garbage after the last datum in the line separated by tabs
            data = data.split("\t")[0]
        if line.find(":") == 6:  # multiline variable
            if medpc_name == "     0":  # first line of multiline variable
                multiline_variable_name = session_lines[i - 1].split(":")[0]
                if multiline_variable_name in medpc_name_to_info_dict:
                    output_name = medpc_name_to_info_dict[multiline_variable_name]["name"]
                    session_dict[output_name] = []
            if multiline_variable_name not in medpc_name_to_info_dict:
                continue
            data = data.split(" ")
            for datum in data:
                datum = datum.strip()
                if datum == "":
                    continue
                output_name = medpc_name_to_info_dict[multiline_variable_name]["name"]
                session_dict[output_name].append(datum)

        # single line variable
        elif medpc_name in medpc_name_to_info_dict:
            output_name = medpc_name_to_info_dict[medpc_name]["name"]
            session_dict[output_name] = data

    # Convert the data types
    for info in medpc_name_to_info_dict.values():
        output_name = info["name"]
        is_array = info["is_array"]
        if output_name in session_dict:
            if is_array:
                if session_dict[output_name] == "":
                    session_dict[output_name] = np.array([], dtype=float)
                elif type(session_dict[output_name]) == str:  # not a multiline variable
                    raise ValueError(
                        f"Expected {output_name} to be a multiline variable, but found a single line variable."
                    )
                else:
                    session_dict[output_name] = np.array(session_dict[output_name], dtype=float)
                    session_dict[output_name] = np.trim_zeros(
                        session_dict[output_name], trim="b"
                    )  # MEDPC adds extra zeros to the end of the array
    return session_dict



from typing import Literal

from pydantic import ConfigDict, FilePath, validate_call

from ..baseimagingextractorinterface import BaseImagingExtractorInterface
from ....utils import ArrayType


class Hdf5ImagingInterface(BaseImagingExtractorInterface):
    """Interface for HDF5 imaging data."""

    display_name = "HDF5 Imaging"
    associated_suffixes = (".h5", ".hdf5")
    info = "Interface for HDF5 imaging data."

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def __init__(
        self,
        file_path: FilePath,
        mov_field: str = "mov",
        sampling_frequency: float = None,
        start_time: float = None,
        metadata: dict = None,
        channel_names: ArrayType = None,
        verbose: bool = True,
        photon_series_type: Literal["OnePhotonSeries", "TwoPhotonSeries"] = "TwoPhotonSeries",
    ):
        """

        Parameters
        ----------
        file_path : FilePath
            Path to .h5 or .hdf5 file.
        mov_field : str, default: 'mov'
        sampling_frequency : float, optional
        start_time : float, optional
        metadata : dict, optional
        channel_names : list of str, optional
        verbose : bool, default: True
        """
        super().__init__(
            file_path=file_path,
            mov_field=mov_field,
            sampling_frequency=sampling_frequency,
            start_time=start_time,
            metadata=metadata,
            channel_names=channel_names,
            verbose=verbose,
            photon_series_type=photon_series_type,
        )



import importlib.util

import numpy as np
from pydantic import FilePath, validate_call
from pynwb import NWBFile

from ...baseextractorinterface import BaseExtractorInterface
from ...tools.nwb_helpers import make_nwbfile_from_metadata
from ...utils import (
    get_metadata_schema_for_icephys,
    get_schema_from_hdmf_class,
    get_schema_from_method_signature,
)


class BaseIcephysInterface(BaseExtractorInterface):
    """Primary class for all intracellular NeoInterfaces."""

    keywords = ("intracellular electrophysiology", "patch clamp", "current clamp")

    ExtractorModuleName = "neo"

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = get_schema_from_method_signature(method=cls.__init__, exclude=[])
        return source_schema

    @validate_call
    def __init__(self, file_paths: list[FilePath]):
        # Check if the ndx_dandi_icephys module is available
        dandi_icephys_spec = importlib.util.find_spec("ndx_dandi_icephys")
        if dandi_icephys_spec is not None:
            from ndx_dandi_icephys import DandiIcephysMetadata

            self.DandiIcephysMetadata = DandiIcephysMetadata

            self.HAVE_NDX_DANDI_ICEPHYS = True
        else:
            self.DandiIcephysMetadata = None
            self.HAVE_NDX_DANDI_ICEPHYS = False

        from ...tools.neo import get_number_of_electrodes, get_number_of_segments

        super().__init__(file_paths=file_paths)

        self.readers_list = list()
        for f in file_paths:
            self.readers_list.append(self.get_extractor()(filename=f))

        self.subset_channels = None
        self.n_segments = get_number_of_segments(neo_reader=self.readers_list[0], block=0)
        self.n_channels = get_number_of_electrodes(neo_reader=self.readers_list[0])

        self._timestamps = None

    def get_metadata_schema(self) -> dict:
        metadata_schema = super().get_metadata_schema()
        if self.DandiIcephysMetadata is not None:
            metadata_schema["properties"]["ndx-dandi-icephys"] = get_schema_from_hdmf_class(self.DandiIcephysMetadata)
        metadata_schema["properties"]["Icephys"] = get_metadata_schema_for_icephys()
        return metadata_schema

    def get_metadata(self) -> dict:
        from ...tools.neo import get_number_of_electrodes

        metadata = super().get_metadata()
        metadata["Icephys"] = dict(
            Device=[dict(name="DeviceIcephys", description="no description")],
            Electrodes=[
                dict(name=f"electrode-{i}", description="no description", device="DeviceIcephys")
                for i in range(get_number_of_electrodes(self.readers_list[0]))
            ],
        )
        return metadata

    def get_original_timestamps(self) -> np.ndarray:
        raise NotImplementedError("Icephys interfaces do not yet support timestamps.")

    def get_timestamps(self) -> np.ndarray:
        raise NotImplementedError("Icephys interfaces do not yet support timestamps.")

    def set_aligned_timestamps(self, aligned_timestamps: np.ndarray):
        raise NotImplementedError("Icephys interfaces do not yet support timestamps.")

    def set_aligned_starting_time(self, aligned_starting_time: float):
        raise NotImplementedError("This icephys interface has not specified the method for aligning starting time.")

    def align_by_interpolation(self, unaligned_timestamps: np.ndarray, aligned_timestamps: np.ndarray):
        raise NotImplementedError("Icephys interfaces do not yet support timestamps.")

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: dict = None,
        icephys_experiment_type: str = "voltage_clamp",
        skip_electrodes: tuple[int] = (),
    ):
        """
        Primary function for converting raw (unprocessed) intracellular data to the NWB standard.

        Parameters
        ----------
        nwbfile : NWBFile
            nwb file to which the recording information is to be added
        metadata : dict, optional
            metadata info for constructing the nwb file (optional).
        icephys_experiment_type : {'voltage_clamp', 'current_clamp', 'izero'}
            Type of icephys recording.
        skip_electrodes : tuple, optional
            Electrode IDs to skip. Defaults to ().
        """
        from ...tools.neo import add_neo_to_nwb

        if nwbfile is None:
            nwbfile = make_nwbfile_from_metadata(metadata)

        if (
            self.HAVE_NDX_DANDI_ICEPHYS
            and "ndx-dandi-icephys" in metadata
            and "DandiIcephysMetadata" not in nwbfile.lab_meta_data
        ):
            nwbfile.add_lab_meta_data(self.DandiIcephysMetadata(**metadata["ndx-dandi-icephys"]))

        for i, reader in enumerate(self.readers_list):
            add_neo_to_nwb(
                neo_reader=reader,
                nwbfile=nwbfile,
                metadata=metadata,
                icephys_experiment_type=metadata["Icephys"]["Sessions"][i]["icephys_experiment_type"],
                stimulus_type=metadata["Icephys"]["Sessions"][i]["stimulus_type"],
                skip_electrodes=skip_electrodes,
            )



from typing import Optional

from pydantic import DirectoryPath

from ..baserecordingextractorinterface import BaseRecordingExtractorInterface
from ....utils import get_schema_from_method_signature


class OpenEphysBinaryRecordingInterface(BaseRecordingExtractorInterface):
    """
    Primary data interface for converting binary OpenEphys data (.dat files).

    Uses :py:class:`~spikeinterface.extractors.OpenEphysBinaryRecordingExtractor`.
    """

    display_name = "OpenEphys Binary Recording"
    associated_suffixes = (".dat", ".oebin", ".npy")
    info = "Interface for converting binary OpenEphys recording data."

    ExtractorName = "OpenEphysBinaryRecordingExtractor"

    @classmethod
    def get_stream_names(cls, folder_path: DirectoryPath) -> list[str]:
        from spikeinterface.extractors import OpenEphysBinaryRecordingExtractor

        stream_names, _ = OpenEphysBinaryRecordingExtractor.get_streams(folder_path=folder_path)
        return stream_names

    @classmethod
    def get_source_schema(cls) -> dict:
        """Compile input schema for the RecordingExtractor."""
        source_schema = get_schema_from_method_signature(
            method=cls.__init__, exclude=["recording_id", "experiment_id", "stub_test"]
        )
        source_schema["properties"]["folder_path"][
            "description"
        ] = "Path to directory containing OpenEphys binary files."
        return source_schema

    def __init__(
        self,
        folder_path: DirectoryPath,
        stream_name: Optional[str] = None,
        block_index: Optional[int] = None,
        stub_test: bool = False,
        verbose: bool = True,
        es_key: str = "ElectricalSeries",
    ):
        """
        Initialize reading of OpenEphys binary recording.

        Parameters
        ----------
        folder_path: FolderPathType
            Path to directory containing OpenEphys binary files.
        stream_name : str, optional
            The name of the recording stream to load; only required if there is more than one stream detected.
            Call `OpenEphysRecordingInterface.get_stream_names(folder_path=...)` to see what streams are available.
        block_index : int, optional, default: None
            The index of the block to extract from the data.
        stub_test : bool, default: False
        verbose : bool, default: True
        es_key : str, default: "ElectricalSeries"
        """
        from ._openephys_utils import _read_settings_xml

        self._xml_root = _read_settings_xml(folder_path)

        available_streams = self.get_stream_names(folder_path=folder_path)
        if len(available_streams) > 1 and stream_name is None:
            raise ValueError(
                "More than one stream is detected! "
                "Please specify which stream you wish to load with the `stream_name` argument. "
                "To see what streams are available, call "
                " `OpenEphysRecordingInterface.get_stream_names(folder_path=...)`."
            )
        if stream_name is not None and stream_name not in available_streams:
            raise ValueError(
                f"The selected stream '{stream_name}' is not in the available streams '{available_streams}'!"
            )

        super().__init__(
            folder_path=folder_path, stream_name=stream_name, block_index=block_index, verbose=verbose, es_key=es_key
        )

        if stub_test:
            self.subset_channels = [0, 1]

    def get_metadata(self) -> dict:
        from ._openephys_utils import _get_session_start_time

        metadata = super().get_metadata()

        session_start_time = _get_session_start_time(element=self._xml_root)
        if session_start_time is not None:
            metadata["NWBFile"].update(session_start_time=session_start_time)
        return metadata



from typing import Literal

from pynwb import NWBFile

from neuroconv.tools import get_package


def add_fiber_photometry_device(
    *,
    nwbfile: NWBFile,
    device_metadata: dict,
    device_type: Literal[
        "OpticalFiber",
        "ExcitationSource",
        "Photodetector",
        "BandOpticalFilter",
        "EdgeOpticalFilter",
        "DichroicMirror",
        "Indicator",
    ],
):
    """Add a photometry device to an NWBFile object."""
    valid_device_types = [
        "OpticalFiber",
        "ExcitationSource",
        "Photodetector",
        "BandOpticalFilter",
        "EdgeOpticalFilter",
        "DichroicMirror",
        "Indicator",
    ]
    assert device_type in valid_device_types, f"device_type must be one of {valid_device_types}"
    ndx_fiber_photometry = get_package("ndx_fiber_photometry")

    device_name = device_metadata["name"]
    if device_name in nwbfile.devices:
        return

    photometry_device = dict(
        OpticalFiber=ndx_fiber_photometry.OpticalFiber,
        ExcitationSource=ndx_fiber_photometry.ExcitationSource,
        Photodetector=ndx_fiber_photometry.Photodetector,
        BandOpticalFilter=ndx_fiber_photometry.BandOpticalFilter,
        EdgeOpticalFilter=ndx_fiber_photometry.EdgeOpticalFilter,
        DichroicMirror=ndx_fiber_photometry.DichroicMirror,
        Indicator=ndx_fiber_photometry.Indicator,
    )[device_type](**device_metadata)

    nwbfile.add_device(photometry_device)



from typing import Literal, Optional, Union

import numpy as np
from pynwb import NWBFile
from pynwb.device import Device
from pynwb.ecephys import ElectricalSeries, ElectrodeGroup

from ...baseextractorinterface import BaseExtractorInterface
from ...utils import (
    DeepDict,
    get_base_schema,
    get_schema_from_hdmf_class,
)


class BaseRecordingExtractorInterface(BaseExtractorInterface):
    """Parent class for all RecordingExtractorInterfaces."""

    keywords = ("extracellular electrophysiology", "voltage", "recording")

    ExtractorModuleName = "spikeinterface.extractors"

    def __init__(self, verbose: bool = True, es_key: str = "ElectricalSeries", **source_data):
        """
        Parameters
        ----------
        verbose : bool, default: True
            If True, will print out additional information.
        es_key : str, default: "ElectricalSeries"
            The key of this ElectricalSeries in the metadata dictionary.
        source_data : dict
            The key-value pairs of extractor-specific arguments.

        """
        super().__init__(**source_data)
        self.recording_extractor = self.get_extractor()(**source_data)
        property_names = self.recording_extractor.get_property_keys()
        # TODO remove this and go and change all the uses of channel_name once spikeinterface > 0.101.0 is released
        if "channel_name" not in property_names and "channel_names" in property_names:
            channel_names = self.recording_extractor.get_property("channel_names")
            self.recording_extractor.set_property("channel_name", channel_names)
            self.recording_extractor.delete_property("channel_names")

        self.subset_channels = None
        self.verbose = verbose
        self.es_key = es_key
        self._number_of_segments = self.recording_extractor.get_num_segments()

    def get_metadata_schema(self) -> dict:
        """Compile metadata schema for the RecordingExtractor."""
        metadata_schema = super().get_metadata_schema()
        metadata_schema["properties"]["Ecephys"] = get_base_schema(tag="Ecephys")
        metadata_schema["properties"]["Ecephys"]["required"] = ["Device", "ElectrodeGroup"]
        metadata_schema["properties"]["Ecephys"]["properties"] = dict(
            Device=dict(type="array", minItems=1, items={"$ref": "#/properties/Ecephys/definitions/Device"}),
            ElectrodeGroup=dict(
                type="array", minItems=1, items={"$ref": "#/properties/Ecephys/definitions/ElectrodeGroup"}
            ),
            Electrodes=dict(
                type="array",
                minItems=0,
                renderForm=False,
                items={"$ref": "#/properties/Ecephys/definitions/Electrodes"},
            ),
        )
        # Schema definition for arrays
        metadata_schema["properties"]["Ecephys"]["definitions"] = dict(
            Device=get_schema_from_hdmf_class(Device),
            ElectrodeGroup=get_schema_from_hdmf_class(ElectrodeGroup),
            Electrodes=dict(
                type="object",
                additionalProperties=False,
                required=["name"],
                properties=dict(
                    name=dict(type="string", description="name of this electrodes column"),
                    description=dict(type="string", description="description of this electrodes column"),
                ),
            ),
        )

        if self.es_key is not None:
            metadata_schema["properties"]["Ecephys"]["properties"].update(
                {self.es_key: get_schema_from_hdmf_class(ElectricalSeries)}
            )
        return metadata_schema

    def get_metadata(self) -> DeepDict:
        metadata = super().get_metadata()

        channel_groups_array = self.recording_extractor.get_channel_groups()
        unique_channel_groups = set(channel_groups_array) if channel_groups_array is not None else ["ElectrodeGroup"]
        electrode_metadata = [
            dict(name=str(group_id), description="no description", location="unknown", device="DeviceEcephys")
            for group_id in unique_channel_groups
        ]

        metadata["Ecephys"] = dict(
            Device=[dict(name="DeviceEcephys", description="no description")],
            ElectrodeGroup=electrode_metadata,
        )

        if self.es_key is not None:
            metadata["Ecephys"][self.es_key] = dict(
                name=self.es_key, description=f"Acquisition traces for the {self.es_key}."
            )

        return metadata

    def get_original_timestamps(self) -> Union[np.ndarray, list[np.ndarray]]:
        """
        Retrieve the original unaltered timestamps for the data in this interface.

        This function should retrieve the data on-demand by re-initializing the IO.

        Returns
        -------
        timestamps: numpy.ndarray or list of numpy.ndarray
            The timestamps for the data stream; if the recording has multiple segments, then a list of timestamps is returned.
        """
        new_recording = self.get_extractor()(
            **{keyword: value for keyword, value in self.source_data.items() if keyword not in ["verbose", "es_key"]}
        )
        if self._number_of_segments == 1:
            return new_recording.get_times()
        else:
            return [
                new_recording.get_times(segment_index=segment_index)
                for segment_index in range(self._number_of_segments)
            ]

    def get_timestamps(self) -> Union[np.ndarray, list[np.ndarray]]:
        """
        Retrieve the timestamps for the data in this interface.

        Returns
        -------
        timestamps: numpy.ndarray or list of numpy.ndarray
            The timestamps for the data stream; if the recording has multiple segments, then a list of timestamps is returned.
        """
        if self._number_of_segments == 1:
            return self.recording_extractor.get_times()
        else:
            return [
                self.recording_extractor.get_times(segment_index=segment_index)
                for segment_index in range(self._number_of_segments)
            ]

    def set_aligned_timestamps(self, aligned_timestamps: np.ndarray):
        assert (
            self._number_of_segments == 1
        ), "This recording has multiple segments; please use 'align_segment_timestamps' instead."

        self.recording_extractor.set_times(times=aligned_timestamps, with_warning=False)

    def set_aligned_segment_timestamps(self, aligned_segment_timestamps: list[np.ndarray]):
        """
        Replace all timestamps for all segments in this interface with those aligned to the common session start time.

        Must be in units seconds relative to the common 'session_start_time'.

        Parameters
        ----------
        aligned_segment_timestamps : list of numpy.ndarray
            The synchronized timestamps for segment of data in this interface.
        """
        assert isinstance(
            aligned_segment_timestamps, list
        ), "Recording has multiple segment! Please pass a list of timestamps to align each segment."
        assert (
            len(aligned_segment_timestamps) == self._number_of_segments
        ), f"The number of timestamp vectors ({len(aligned_segment_timestamps)}) does not match the number of segments ({self._number_of_segments})!"

        for segment_index in range(self._number_of_segments):
            self.recording_extractor.set_times(
                times=aligned_segment_timestamps[segment_index],
                segment_index=segment_index,
                with_warning=False,
            )

    def set_aligned_starting_time(self, aligned_starting_time: float):
        if self._number_of_segments == 1:
            self.set_aligned_timestamps(aligned_timestamps=self.get_timestamps() + aligned_starting_time)
        else:
            self.set_aligned_segment_timestamps(
                aligned_segment_timestamps=[
                    segment_timestamps + aligned_starting_time for segment_timestamps in self.get_timestamps()
                ]
            )

    def set_aligned_segment_starting_times(self, aligned_segment_starting_times: list[float]):
        """
        Align the starting time for each segment in this interface relative to the common session start time.

        Must be in units seconds relative to the common 'session_start_time'.

        Parameters
        ----------
        aligned_segment_starting_times : list of floats
            The starting time for each segment of data in this interface.
        """
        assert len(aligned_segment_starting_times) == self._number_of_segments, (
            f"The length of the starting_times ({len(aligned_segment_starting_times)}) does not match the "
            "number of segments ({self._number_of_segments})!"
        )

        if self._number_of_segments == 1:
            self.set_aligned_starting_time(aligned_starting_time=aligned_segment_starting_times[0])
        else:
            aligned_segment_timestamps = [
                segment_timestamps + aligned_segment_starting_time
                for segment_timestamps, aligned_segment_starting_time in zip(
                    self.get_timestamps(), aligned_segment_starting_times
                )
            ]
            self.set_aligned_segment_timestamps(aligned_segment_timestamps=aligned_segment_timestamps)

    def set_probe(self, probe, group_mode: Literal["by_shank", "by_probe"]):
        """
        Set the probe information via a ProbeInterface object.

        Parameters
        ----------
        probe : probeinterface.Probe
            The probe object.
        group_mode : {'by_shank', 'by_probe'}
            How to group the channels. If 'by_shank', channels are grouped by the shank_id column.
            If 'by_probe', channels are grouped by the probe_id column.
            This is a required parameter to avoid the pitfall of using the wrong mode.
        """
        # Set the probe to the recording extractor
        self.recording_extractor.set_probe(
            probe,
            in_place=True,
            group_mode=group_mode,
        )
        # Spike interface sets the "group" property
        # But neuroconv allows "group_name" property to override spike interface "group" value
        self.recording_extractor.set_property("group_name", self.recording_extractor.get_property("group").astype(str))

    def has_probe(self) -> bool:
        """
        Check if the recording extractor has probe information.

        Returns
        -------
        has_probe : bool
            True if the recording extractor has probe information.
        """
        return self.recording_extractor.has_probe()

    def align_by_interpolation(
        self,
        unaligned_timestamps: np.ndarray,
        aligned_timestamps: np.ndarray,
    ):
        if self._number_of_segments == 1:
            self.set_aligned_timestamps(
                aligned_timestamps=np.interp(x=self.get_timestamps(), xp=unaligned_timestamps, fp=aligned_timestamps)
            )
        else:
            raise NotImplementedError("Multi-segment support for aligning by interpolation has not been added yet.")

    def subset_recording(self, stub_test: bool = False):
        """
        Subset a recording extractor according to stub and channel subset options.

        Parameters
        ----------
        stub_test : bool, default: False
        """
        from spikeinterface.core.segmentutils import AppendSegmentRecording

        max_frames = 100

        recording_extractor = self.recording_extractor
        number_of_segments = recording_extractor.get_num_segments()
        recording_segments = [recording_extractor.select_segments([index]) for index in range(number_of_segments)]
        end_frame_list = [min(max_frames, segment.get_num_frames()) for segment in recording_segments]
        recording_segments_stubbed = [
            segment.frame_slice(start_frame=0, end_frame=end_frame)
            for segment, end_frame in zip(recording_segments, end_frame_list)
        ]
        recording_extractor_stubbed = AppendSegmentRecording(recording_list=recording_segments_stubbed)

        times_stubbed = [
            recording_extractor.get_times(segment_index=segment_index)[:end_frame]
            for segment_index, end_frame in zip(range(number_of_segments), end_frame_list)
        ]
        for segment_index in range(number_of_segments):
            recording_extractor_stubbed.set_times(
                times=times_stubbed[segment_index],
                segment_index=segment_index,
                with_warning=False,
            )

        return recording_extractor_stubbed

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: Optional[dict] = None,
        stub_test: bool = False,
        starting_time: Optional[float] = None,
        write_as: Literal["raw", "lfp", "processed"] = "raw",
        write_electrical_series: bool = True,
        compression: Optional[str] = None,  # TODO: remove completely after 10/1/2024
        compression_opts: Optional[int] = None,
        iterator_type: Optional[str] = "v2",
        iterator_opts: Optional[dict] = None,
    ):
        """
        Primary function for converting raw (unprocessed) RecordingExtractor data to the NWB standard.

        Parameters
        ----------
        nwbfile : NWBFile
            NWBFile to which the recording information is to be added
        metadata : dict, optional
            metadata info for constructing the NWB file.
            Should be of the format::

                metadata['Ecephys']['ElectricalSeries'] = dict(name=my_name, description=my_description)

        starting_time : float, optional
            Sets the starting time of the ElectricalSeries to a manually set value.
        stub_test : bool, default: False
            If True, will truncate the data to run the conversion faster and take up less memory.
        write_as : {'raw', 'lfp', 'processed'}
        write_electrical_series : bool, default: True
            Electrical series are written in acquisition. If False, only device, electrode_groups,
            and electrodes are written to NWB.
        iterator_type : {'v2'}
            The type of DataChunkIterator to use.
            'v2' is the locally developed RecordingExtractorDataChunkIterator, which offers full control over chunking
        iterator_opts : dict, optional
            Dictionary of options for the RecordingExtractorDataChunkIterator (iterator_type='v2').
            Valid options are:

            * buffer_gb : float, default: 1.0
                In units of GB. Recommended to be as much free RAM as available. Automatically calculates suitable
                buffer shape.
            * buffer_shape : tuple, optional
                Manual specification of buffer shape to return on each iteration.
                Must be a multiple of chunk_shape along each axis.
                Cannot be set if `buffer_gb` is specified.
            * chunk_mb : float. default: 1.0
                Should be below 1 MB. Automatically calculates suitable chunk shape.
            * chunk_shape : tuple, optional
                Manual specification of the internal chunk shape for the HDF5 dataset.
                Cannot be set if `chunk_mb` is also specified.
            * display_progress : bool, default: False
                Display a progress bar with iteration rate and estimated completion time.
            * progress_bar_options : dict, optional
                Dictionary of keyword arguments to be passed directly to tqdm.
                See https://github.com/tqdm/tqdm#parameters for options.
        """
        from ...tools.spikeinterface import add_recording_to_nwbfile

        if stub_test or self.subset_channels is not None:
            recording = self.subset_recording(stub_test=stub_test)
        else:
            recording = self.recording_extractor

        if metadata is None:
            metadata = self.get_metadata()

        add_recording_to_nwbfile(
            recording=recording,
            nwbfile=nwbfile,
            metadata=metadata,
            starting_time=starting_time,
            write_as=write_as,
            write_electrical_series=write_electrical_series,
            es_key=self.es_key,
            compression=compression,
            compression_opts=compression_opts,
            iterator_type=iterator_type,
            iterator_opts=iterator_opts,
        )



from .checks import calculate_regular_series_rate
from .dict import (
    DeepDict,
    append_replace_dict_in_list,
    dict_deep_update,
    exist_dict_in_list,
    load_dict_from_file,
)
from .json_schema import (
    _NWBMetaDataEncoder,
    fill_defaults,
    get_base_schema,
    get_metadata_schema_for_icephys,
    get_schema_from_hdmf_class,
    get_schema_from_method_signature,
    unroot_schema,
    get_json_schema_from_method_signature,
)
from .types import (
    ArrayType,
    IntType,
    OptionalArrayType,
)


# TODO: remove after 3/1/2025
def __getattr__(name):
    from warnings import warn
    from typing import Optional

    from pydantic import FilePath, DirectoryPath

    if name == "FilePathType":
        message = (
            "The 'FilePathType' type has been deprecated and will be removed after 3/1/2025. "
            "Please use `pydantic.FilePath` instead."
        )
        warn(message=message, category=DeprecationWarning, stacklevel=2)

        return FilePath
    if name == "OptionalFilePathType":
        message = (
            "The 'OptionalFilePathType' type has been deprecated and will be removed after 3/1/2025. "
            "Please use `typing.Optional[pydantic.FilePath]` instead."
        )
        warn(message=message, category=DeprecationWarning, stacklevel=2)

        return Optional[FilePath]
    if name == "FolderPathType":
        message = (
            "The 'FolderPathType' type has been deprecated and will be removed after 3/1/2025. "
            "Please use `pydantic.DirectoryPath` instead."
        )
        warn(message=message, category=DeprecationWarning, stacklevel=2)

        return DirectoryPath
    if name == "OptionalFolderPathType":
        message = (
            "The 'OptionalFolderPathType' type has been deprecated and will be removed after 3/1/2025. "
            "Please use `typing.Optional[pydantic.DirectoryPath]` instead."
        )
        warn(message=message, category=DeprecationWarning, stacklevel=2)

        return Optional[DirectoryPath]



import sys
from importlib import import_module
from pathlib import Path
from typing import Optional

import click
from jsonschema import RefResolver, validate
from pydantic import DirectoryPath, FilePath

from ...nwbconverter import NWBConverter
from ...utils import dict_deep_update, load_dict_from_file


@click.command()
@click.argument("specification-file-path")
@click.option(
    "--data-folder-path",
    help="Path to folder where the source data may be found.",
    type=click.Path(writable=True),
)
@click.option(
    "--output-folder-path",
    default=None,
    help="Path to folder where you want to save the output NWBFile.",
    type=click.Path(writable=True),
)
@click.option("--overwrite", help="Overwrite an existing NWBFile at the location.", is_flag=True)
def run_conversion_from_yaml_cli(
    specification_file_path: str,
    data_folder_path: Optional[str] = None,
    output_folder_path: Optional[str] = None,
    overwrite: bool = False,
):
    """
    Run the tool function 'run_conversion_from_yaml' via the command line.

    specification-file-path :
    Path to the .yml specification file.
    """
    run_conversion_from_yaml(
        specification_file_path=specification_file_path,
        data_folder_path=data_folder_path,
        output_folder_path=output_folder_path,
        overwrite=overwrite,
    )


def run_conversion_from_yaml(
    specification_file_path: FilePath,
    data_folder_path: Optional[DirectoryPath] = None,
    output_folder_path: Optional[DirectoryPath] = None,
    overwrite: bool = False,
):
    """
    Run conversion to NWB given a yaml specification file.

    Parameters
    ----------
    specification_file_path : FilePathType
        File path leading to .yml specification file for NWB conversion.
    data_folder_path : FolderPathType, optional
        Folder path leading to root location of the data files.
        The default is the parent directory of the specification_file_path.
    output_folder_path : FolderPathType, optional
        Folder path leading to the desired output location of the .nwb files.
        The default is the parent directory of the specification_file_path.
    overwrite : bool, default: False
        If True, replaces any existing NWBFile at the nwbfile_path location, if save_to_file is True.
        If False, appends the existing NWBFile at the nwbfile_path location, if save_to_file is True.
    """
    from dandi.organize import create_unique_filenames_from_metadata
    from dandi.pynwb_utils import _get_pynwb_metadata

    if data_folder_path is None:
        data_folder_path = Path(specification_file_path).parent
    if output_folder_path is None:
        output_folder_path = Path(specification_file_path).parent
    else:
        output_folder_path = Path(output_folder_path)
    specification = load_dict_from_file(file_path=specification_file_path)
    schema_folder = Path(__file__).parent.parent.parent / "schemas"
    specification_schema = load_dict_from_file(file_path=schema_folder / "yaml_conversion_specification_schema.json")
    sys_uri_base = "file:/" if sys.platform.startswith("win32") else "file://"
    validate(
        instance=specification,
        schema=specification_schema,
        resolver=RefResolver(base_uri=sys_uri_base + str(schema_folder) + "/", referrer=specification_schema),
    )

    global_metadata = specification.get("metadata", dict())
    global_conversion_options = specification.get("conversion_options", dict())
    data_interfaces_spec = specification.get("data_interfaces")
    data_interfaces_module = import_module(name=".datainterfaces", package="neuroconv")
    data_interface_classes = {key: getattr(data_interfaces_module, name) for key, name in data_interfaces_spec.items()}

    CustomNWBConverter = type(
        "CustomNWBConverter", (NWBConverter,), dict(data_interface_classes=data_interface_classes)
    )

    file_counter = 0
    for experiment in specification["experiments"].values():
        experiment_metadata = experiment.get("metadata", dict())
        for session in experiment["sessions"]:
            file_counter += 1
            source_data = session["source_data"]
            for interface_name, interface_source_data in session["source_data"].items():
                for key, value in interface_source_data.items():
                    if key == "file_paths":
                        source_data[interface_name].update({key: [str(Path(data_folder_path) / x) for x in value]})
                    elif key in ("file_path", "folder_path"):
                        source_data[interface_name].update({key: str(Path(data_folder_path) / value)})
            converter = CustomNWBConverter(source_data=source_data)
            metadata = converter.get_metadata()
            for metadata_source in [global_metadata, experiment_metadata, session.get("metadata", dict())]:
                metadata = dict_deep_update(metadata, metadata_source)
            nwbfile_name = session.get("nwbfile_name", f"temp_nwbfile_name_{file_counter}").strip(".nwb")
            session_conversion_options = session.get("conversion_options", dict())
            conversion_options = dict()
            for key in converter.data_interface_objects:
                conversion_options[key] = dict(session_conversion_options.get(key, dict()), **global_conversion_options)
            converter.run_conversion(
                nwbfile_path=output_folder_path / f"{nwbfile_name}.nwb",
                metadata=metadata,
                overwrite=overwrite,
                conversion_options=conversion_options,
            )
    # To properly mimic a true dandi organization, the full directory must be populated with NWBFiles.
    all_nwbfile_paths = [nwbfile_path for nwbfile_path in output_folder_path.iterdir() if nwbfile_path.suffix == ".nwb"]
    nwbfile_paths_to_set = [
        nwbfile_path for nwbfile_path in all_nwbfile_paths if "temp_nwbfile_name_" in nwbfile_path.stem
    ]
    if any(nwbfile_paths_to_set):
        dandi_metadata_list = list()
        for nwbfile_path_to_set in nwbfile_paths_to_set:
            dandi_metadata = _get_pynwb_metadata(path=nwbfile_path_to_set)
            dandi_metadata.update(path=nwbfile_path_to_set)
            dandi_metadata_list.append(dandi_metadata)
        dandi_metadata_with_set_paths = create_unique_filenames_from_metadata(metadata=dandi_metadata_list)

        for nwbfile_path_to_set, dandi_metadata_with_set_path in zip(
            nwbfile_paths_to_set, dandi_metadata_with_set_paths
        ):
            dandi_filename = dandi_metadata_with_set_path["dandi_filename"]

            assert (
                dandi_filename != ".nwb"
            ), f"Not enough metadata available to assign name to {str(nwbfile_path_to_set)}!"

            # Rename file on system
            nwbfile_path_to_set.rename(str(output_folder_path / dandi_filename))



"""Utilities used by the SpikeGLX interfaces."""

import json
from datetime import datetime
from pathlib import Path

from pydantic import FilePath


def add_recording_extractor_properties(recording_extractor) -> None:
    """Utility functions for setting some properties on the recording extractor"""
    probe = recording_extractor.get_probe()
    channel_ids = recording_extractor.get_channel_ids()

    # Should follow pattern 'Imec0', 'Imec1', etc.
    probe_name = recording_extractor.stream_id[:5].capitalize()

    if probe.get_shank_count() > 1:
        shank_ids = probe.shank_ids
        recording_extractor.set_property(key="shank_ids", values=shank_ids)
        group_name = [f"{probe_name}{shank_id}" for shank_id in shank_ids]
    else:
        group_name = [f"{probe_name}"] * len(channel_ids)

    recording_extractor.set_property(key="group_name", ids=channel_ids, values=group_name)

    contact_shapes = probe.contact_shapes  # The geometry of the contact shapes
    recording_extractor.set_property(key="contact_shapes", ids=channel_ids, values=contact_shapes)

    contact_ids = probe.contact_ids  # s{shank_number}e{electrode_number} or e{electrode_number}
    recording_extractor.set_property(key="contact_ids", ids=channel_ids, values=contact_ids)


def get_session_start_time(recording_metadata: dict) -> datetime:
    """
    Fetches the session start time from the recording_metadata dictionary.

    Parameters
    ----------
    recording_metadata : dict
        The metadata dictionary as obtained from the Spikelgx recording.

    Returns
    -------
    datetime or None
        the session start time in datetime format.
    """
    session_start_time = recording_metadata.get("fileCreateTime", None)
    if session_start_time.startswith("0000-00-00"):
        # date was removed. This sometimes happens with human data to protect the
        # anonymity of medical patients.
        return
    if session_start_time:
        session_start_time = datetime.fromisoformat(session_start_time)
    return session_start_time


def fetch_stream_id_for_spikelgx_file(file_path: FilePath) -> str:
    """
    Returns the stream_id for a spikelgx file.

    Example of file name structure:
    Consider the filenames: `Noise4Sam_g0_t0.nidq.bin` or `Noise4Sam_g0_t0.imec0.lf.bin`
    The filenames consist of 3 or 4 parts separated by `.`
      1. "Noise4Sam_g0_t0" will be the `name` variable. This chosen by the user at recording time.
      2. "_gt0_" will give the `seg_index` (here 0)
      3. "nidq" or "imec0" will give the `device` variable
      4. "lf" or "ap" will be the `signal_kind` variable (for nidq the signal kind is an empty string)

    stream_id is the concatenation of `device.signal_kind`

    Parameters
    ----------
    file_path : FilePathType
        The file_path of spikelgx file.

    Returns
    -------
    str
        the stream_id
    """
    suffixes = Path(file_path).suffixes
    device = next(suffix for suffix in suffixes if "imec" in suffix or "nidq" in suffix)
    signal_kind = ""
    if "imec" in device:
        signal_kind = next(suffix for suffix in suffixes if "ap" in suffix or "lf" in suffix)

    stream_id = device[1:] + signal_kind

    return stream_id


def get_device_metadata(meta) -> dict:
    """Returns a device with description including the metadata as described here
    # https://billkarsh.github.io/SpikeGLX/Sgl_help/Metadata_30.html

    Returns
    -------
    dict
        a dict containing the metadata necessary for creating the device
    """
    # TODO, get probe metadata from spikeinterface
    metadata_dict = dict()
    if "imDatPrb_type" in meta:
        probe_type_to_probe_description = {
            "0": "NP1.0",
            "21": "NP2.0(1-shank)",
            "24": "NP2.0(4-shank)",
            "1030": "NP1.0 NHP",
        }
        probe_type = str(meta["imDatPrb_type"])
        probe_type_description = probe_type_to_probe_description.get(probe_type, "Unknown SpikeGLX probe type.")
        metadata_dict.update(probe_type=probe_type, probe_type_description=probe_type_description)

    if "imDatFx_pn" in meta:
        metadata_dict.update(flex_part_number=meta["imDatFx_pn"])

    if "imDatBsc_pn" in meta:
        metadata_dict.update(connected_base_station_part_number=meta["imDatBsc_pn"])

    description_string = "A Neuropixel probe of unknown subtype."
    if metadata_dict:
        description_string = json.dumps(metadata_dict)
    device_metadata = dict(name="NeuropixelImec", description=description_string, manufacturer="Imec")

    return device_metadata
